"""Various utilities/helpers for NIMO modem interaction and debugging.
"""
import os
from datetime import datetime, timezone
from typing import Generator


def dprint(raw_string: str) -> str:
    """Returns a printable string on single line."""
    return raw_string.replace('\r', '<cr>').replace('\n', '<lf>')


def vlog(tag: str = 'NONE') -> bool:
    """Returns True if the tag is in the LOG_VERBOSE environment variable."""
    return tag in str(os.getenv('LOG_VERBOSE'))


def ts_to_iso(timestamp: 'float|int', ms: bool = False) -> str:
    """Converts a unix timestamp to ISO 8601 format (UTC).
    
    Args:
        timestamp: A unix timestamp.
        ms: Flag indicating whether to include milliseconds in response
    
    Returns:
        ISO 8601 UTC format e.g. `YYYY-MM-DDThh:mm:ss[.sss]Z`

    """
    iso_time = datetime.fromtimestamp(timestamp, tz=timezone.utc).isoformat()
    if not ms:
        return f'{iso_time[:19]}Z'
    return f'{iso_time[:23]}Z'


def iso_to_ts(iso_time: str, ms: bool = False) -> int:
    """Converts a ISO 8601 timestamp (UTC) to unix timestamp.
    
    Args:
        iso_time: An ISO 8601 UTC datetime `YYYY-MM-DDThh:mm:ss[.sss]Z`
        ms: Flag indicating whether to include milliseconds in response
    
    Returns:
        Unix UTC timestamp as an integer, or float if `ms` flag is set.

    """
    if '.' not in iso_time:
        iso_time = iso_time.replace('Z', '.000Z')
    utc_dt = datetime.strptime(iso_time, '%Y-%m-%dT%H:%M:%S.%fZ')
    ts = (utc_dt - datetime(1970, 1, 1)).total_seconds()
    if not ms:
        ts = int(ts)
    return ts


def bits_in_bitmask(bitmask: int) -> Generator[int]:
    """Yield the integer value of each bit in a bitmask."""
    while bitmask:
        bit = bitmask & (~bitmask+1)
        yield bit
        bitmask ^= bit



# -*- coding: utf-8 -*-
"""NIMO modem constants.

This module provides mapping of constants used within a NIMO modem.

"""
# from dataclasses import dataclass
from enum import Enum, IntEnum, IntFlag

MSG_MO_MAX_SIZE = 6400      # IsatData Pro
MSG_MT_MAX_SIZE = 10000     # IsatData Pro (non-low power)
MSG_MO_NAME_MAX_LEN = 8     # Max characters for name in Orbcomm modems
MSG_MO_NAME_QMAX_LEN = 12   # Max characters for name in Quectel modems
BAUDRATES = [1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200]
GEOSTATIONARY_DISTANCE_M = 35786000


class NimoIntEnum(IntEnum):
    """IntEnum class wrapper with is_valid method."""
    @classmethod
    def is_valid(cls, value) -> bool:
        """True if the value exists in the enumeration."""
        return value in set(item.value for item in cls)


class NimoFloatEnum(Enum):
    """Enum class wrapper with is_valid method."""
    @classmethod
    def is_valid(cls, value) -> bool:
        """True if the value exists in the enumeration."""
        return value in set(item.value for item in cls)


class AtParsingState(NimoIntEnum):
    """Maps AT command response parsing states."""
    ECHO = 0
    RESPONSE = 1
    CRC = 2
    OK = 3
    ERROR = 4


class MessagePriority(NimoIntEnum):
    """Message priorities for NIMO modem messages."""
    NONE = 0
    HIGH = 1
    MEDH = 2
    MEDL = 3
    LOW = 4


class DataFormat(NimoIntEnum):
    """Data formats used for submitting or extracting message data/payload."""
    TEXT = 1
    HEX = 2
    BASE64 = 3


class ControlState(NimoIntEnum):
    """States of the NIMO modem internal network acquisition process."""
    STOPPED = 0
    GNSS_WAIT = 1
    SEARCH_START = 2
    BEAM_SEARCH = 3
    BEAM_FOUND = 4
    BEAM_ACQUIRED = 5
    BEAM_SWITCH = 6
    REGISTERING = 7
    RECEIVE_ONLY = 8
    BB_DOWNLOAD = 9
    ACTIVE = 10
    BLOCKED = 11
    CONFIRM_PREVIOUS_BEAM = 12
    CONFIRM_REQUESTED_BEAM = 13
    CONNECT_CONFIRMED_BEAM = 14


class BeamState(NimoIntEnum):
    """States of the NIMO modem satellite beam internal selection process."""
    IDLE = 0
    SEARCH_ANY_TRAFFIC = 1
    SEARCH_LAST_TRAFFIC = 2
    RESERVED = 3
    SEARCH_NEW_TRAFFIC = 4
    SEARCH_BULLETIN_BOARD = 5
    DELAY_TRAFFIC_SEARCH = 6


class MessageState(NimoIntEnum):
    """Message states of NIMO modem messages."""
    UNAVAILABLE = 0
    RX_PENDING = 1
    RX_COMPLETE = 2
    RX_RETRIEVED = 3
    TX_READY = 4
    TX_SENDING = 5
    TX_COMPLETE = 6
    TX_FAILED = 7
    TX_CANCELLED = 8


class AtErrorCode(NimoIntEnum):
    """AT command error codes for NIMO modems."""
    # Standard / documented
    OK = 0
    ERROR = 4
    INVALID_CRC = 100
    UNKNOWN_COMMAND = 101
    INVALID_COMMAND_PARAMETERS = 102
    MESSAGE_LENGTH_EXCEEDS_FORMAT_SIZE = 103
    RESERVED_104 = 104
    SYSTEM_ERROR = 105
    QUEUE_INSUFFICIENT_RESOURCES = 106
    DUPLICATE_MESSAGE_NAME = 107
    GNSS_TIMEOUT = 108
    MESSAGE_UNAVAILABLE = 109
    RESERVED_110 = 110
    RESERVED_111 = 111
    READ_ONLY_PARAMETER = 112
    # Extensions for additional situations
    TIMEOUT = 255
    CRC_CONFIG_MISMATCH = 254
    UNABLE_TO_DELETE = 253


class PowerMode(NimoIntEnum):
    """The Power Mode setting of the NIMO modem.
    
    Implies various internal state machine settings for balancing power
    consumption against speed of recovery from line of sight blockages.
    
    """
    MOBILE_POWERED = 0
    FIXED_POWERED = 1
    MOBILE_BATTERY = 2
    FIXED_BATTERY = 3
    MOBILE_MINIMAL = 4
    MOBILE_PARKED = 5

    def gnss_refresh_hours(self):
        """The minimum GNSS refresh interval [hours]."""
        if self.value == 0:
            return 3
        if self.value == 1:
            return 24
        if self.value == 2:
            return 6
        if self.value == 3:
            return 14 * 24
        if self.value == 4:
            return 12
        if self.value == 5:
            return 24
    
    def transmit_lifetime_seconds(self):
        """The maximum duration of a message in the transmit queue [seconds]."""
        if self.value in [0, 1]:
            return 3 * 3600
        return  3 * 60
    
    def beam_search_interval_seconds(self):
        """The minimum time between background beam searches [seconds]."""
        if self.value in [0, 2, 4]:
            return 20 * 60
        return 60 * 60

    def short_term_blockage_seconds(self):
        """The time in blockage before initiating a beam search [seconds]."""
        if self.value in [0, 1]:
            return 5 * 60
        if self.value in [2, 3, 4]:
            return 20 * 60
        return 15 * 60
    
    def beam_search_maximum_seconds(self):
        """The maximum backoff interval between searches when blocked [seconds].
        """
        if self.value in [0, 1]:
            return 0
        return 1600 * 60


class WakeupPeriod(NimoIntEnum):
    """The Wakeup Period setting of a NIMO modem.
    
    Determines how often the modem wakes up to listen briefly for potential
    mobile-terminated messages to be delivered by the network.
    
    """
    NONE = 0   # 5 seconds
    SECONDS_30 = 1
    MINUTES_1 = 2
    MINUTES_3 = 3
    MINUTES_10 = 4
    MINUTES_30 = 5
    MINUTES_60 = 6
    MINUTES_2 = 7
    MINUTES_5 = 8
    MINUTES_15 = 9
    MINUTES_20 = 10

    def seconds(self):
        if self.name == 'NONE':
            return 5
        value = int(self.name.split('_'))[1]
        if self.name.startswith('MINUTES'):
            return value * 60
        return value


class WakeupWay(NimoIntEnum):
    """Quectel CC200A-LB wakeup methods."""
    WAKEUP_PIN = 0
    UART = 1


class WorkMode(NimoIntEnum):
    """Quectel CC200A-LB working modes."""
    WORKING = 1
    GNSS = 2
    PERIODIC_SLEEP = 3


class UrcControl(IntFlag):
    """Control bits for Quectel Unsolicited Response Codes."""
    GNSS_FIX_NEW =              0b00000001
    MESSAGE_MT_RECEIVED =       0b00000010
    MESSAGE_MO_COMPLETE =       0b00000100
    NETWORK_REGISTERED =        0b00001000
    WAKEUP_PERIOD_CHANGE =      0b00010000
    UTC_TIME_SYNC =             0b00100000
    GNSS_FIX_TIMEOUT =          0b01000000
    NETWORK_PING_ACKNOWLEDGED = 0b10000000
    
    @classmethod
    def get_events(cls, event_mask: int) -> 'list[EventNotification]':
        """Parses a bitmask to return a list of events."""
        return [item for item in cls if item.value & event_mask]


class UrcCode(NimoIntEnum):
    """Quectel URC code map."""
    GPS_FIX = 0
    RX_END = 1
    TX_END = 2
    REGED = 3
    ITV_CHG = 4
    TIME_UPD = 5
    GPS_TMO = 6
    PLG_RESP = 7


class GnssMode(NimoIntEnum):
    """Base class for manufacturer-specific variants."""


class GnssModeOrbcomm(GnssMode):
    """The operating mode setting for the built-in GNSS in a NIMO modem."""
    GPS = 0
    GLONASS = 1
    BEIDOU = 2
    GALILEO = 3
    GPS_GLONASS = 10
    GPS_BEIDOU = 11
    GLONASS_BEIDOU = 12
    GPS_GALILEO = 13
    GLONASS_GALILEO = 14
    BEIDOU_GALILEO = 15


class GnssModeQuectel(GnssMode):
    GPS = 0
    GPS_BDS = 1
    GPS_GLONASS = 2
    GPS_GALILEO = 3
    GPS_GLONASS_GALILEO_BDS = 4


class GnssDynamicPlatformModel(NimoIntEnum):
    """The dynamic acquisition and tracking model used by the modem's GNSS.
    
    SUPPORTED ON SELECT MODEMS ONLY. CONSULT MANUFACTURER PRIOR TO USE.
    
    """
    PORTABLE = 0
    STATIONARY = 2
    PEDESTRIAN = 3
    AUTOMOTIVE = 4
    SEA = 5
    AIR_1G = 6
    AIR_2G = 7
    AIR_4G = 8


class SignalLevelRegional(NimoFloatEnum):
    """Qualitative descriptors for SNR/CN0 values for a NIMO Regional Beam.
    
    BARS_n: *n* is a scale from 0..5 to be used as greaterThan threshold
    NONE, MARGINAL, GOOD: a scale to be used as greaterOrEqual threshold

    """
    BARS_0 = 0
    BARS_1 = 37.0
    BARS_2 = 39.0
    BARS_3 = 41.0
    BARS_4 = 43.0
    BARS_5 = 45.5
    INVALID = 55.0


class SignalQuality(NimoIntEnum):
    """Qualitative descriptor corresponding to a SignalLevel"""
    NONE = 0
    WEAK = 1
    LOW = 2
    MID = 3
    GOOD = 4
    STRONG = 5
    WARNING = 6


class EventNotification(IntFlag):
    """Bitmask enumerated values for NIMO modem events."""
    GNSS_FIX_NEW =              0b000000000001
    MESSAGE_MT_RECEIVED =       0b000000000010
    MESSAGE_MO_COMPLETE =       0b000000000100
    NETWORK_REGISTERED =        0b000000001000
    MODEM_RESET_COMPLETE =      0b000000010000
    JAMMING_ANTENNA_CHANGE =    0b000000100000
    MODEM_RESET_PENDING =       0b000001000000
    WAKEUP_PERIOD_CHANGE =      0b000010000000
    UTC_TIME_SYNC =             0b000100000000
    GNSS_FIX_TIMEOUT =          0b001000000000
    EVENT_TRACE_CACHED =        0b010000000000
    NETWORK_PING_ACKNOWLEDGED = 0b100000000000
    
    @classmethod
    def get_events(cls, event_mask: int) -> 'list[EventNotification]':
        """Parses a bitmask to return a list of events."""
        return [item for item in cls if item.value & event_mask]


class NetworkStatus(NimoIntEnum):
    """Simplified state of network acquisition and tracking."""
    UNKNOWN = 0
    RX_STOPPED = 1
    RX_SEARCHING = 2
    RX_ACQUIRING = 3
    RX_ONLY_NOT_REGISTERED = 4
    OK = 5
    SUSPENDED = 6
    MUTED = 7
    BLOCKED = 8


class EventTraceClass(NimoIntEnum):
    """Event Trace categories for the NIMO modem."""
    HARDWARE_FAULT = 1
    SYSTEM = 2
    SATELLITE = 3
    GNSS = 4
    MESSAGE = 5


class EventTraceSubclass(NimoIntEnum):
    """Base class for trace subclasses"""


class EventTraceSystem(EventTraceSubclass):
    """Event Trace subcategories for NIMO modem System events."""
    RESET = 1
    LOW_POWER = 2
    SYSTEM_STATS = 3
    SATCOM_STATS = 4


class EventTraceSatellite(EventTraceSubclass):
    """Event Trace subcategories for NIMO modem Satellite events."""
    RX_STATUS = 1
    TX_STATUS = 2
    BEAM_CONNECT = 3
    BEAM_SEARCH_RESULT = 4
    GEO_ADJUST = 5
    EVENT_LOG = 6
    RX_METRICS_SUBFRAME = 16
    RX_METRICS_MINUTE = 17
    RX_METRICS_HOUR = 18
    RX_METRICS_DAY = 19
    TX_METRICS_SUBFRAME = 20
    TX_METRICS_MINUTE = 21
    TX_METRICS_HOUR = 22
    TX_METRICS_DAY = 23


class EventTraceGnss(EventTraceSubclass):
    """Event Trace subcategories for NIMO modem GNSS events."""
    FIX_STATS = 1
    DOPPLER = 4


class EventTraceMessage(EventTraceSubclass):
    """Event Trace subcategories for NIMO modem Message events."""
    RECEIVE_STATS = 1
    TRANSMIT_STATS = 2
    TRANSMIT_UTILITY = 3


class EventTrace:
    def __init__(self,
                 trace_class: EventTraceClass,
                 trace_subclass: EventTraceSubclass,
                 data: 'tuple[str, str|dict|IntEnum]') -> None:
        """Defines an event trace.
        
        Args:
            trace_class: The enumerated class
            trace_subclass: The enumerated subclass dependent on the class
            data: The set of (name, meta) where meta defines either a
                data type (e.g. `uint`) or a mapping (`dict` or `IntEnum`)
        """
        self.trace_class: EventTraceClass = trace_class
        self.trace_subclass: EventTraceSubclass = trace_subclass
        self.data: 'tuple[str, str|dict|IntEnum]' = data


EVENT_TRACE_SATELLITE_GENERAL = EventTrace(
    trace_class=EventTraceClass.SATELLITE,
    trace_subclass=EventTraceSatellite.RX_STATUS,
    data=(
        ('subframe_number', 'uint'),
        ('traffic_vcid', 'uint'),
        ('configuration_id', 'uint'),
        ('beam_number', 'uint'),
        ('reserved04', 'uint'),
        ('tx_access_sip', 'uint'),
        ('tx_access_operator', 'uint'),
        ('tx_access_user', 'uint'),
        ('tx_suspend_flags', {
            0x1: 'BEAM_REGISTRATION',
            0x2: 'BEAM_SWITCH',
            0x4: 'RESERVED',
            0x8: 'BLOCKED',
        }),
        ('tx_messages_active', 'uint'),
        ('tx_messages_total', 'uint'),
        ('tx_packets_active', 'uint'),
        ('tx_state', {
            0: 'ACTIVE',
            1: 'SUSPENDING',
            2: 'SUSPENDED_PENDING_GNSS'}),
        ('active_rx_messages', 'uint'),
        ('beamswitch_averaging_window', 'uint'),
        ('beamswitch_averaging_count', 'uint'),
        ('c_n_x100', 'uint'),
        ('beamsample_threshold', 'uint'),
        ('beamsample_timer', 'uint'),
        ('flags', {
            0x1: 'REGISTERED',
            0x2: 'SENDING_BEAM_REGISTRATION',
            0x10: 'BEAM_SEARCH',
            0x20: 'BEAM_SAMPLE_REQUIRED',
            0x40: 'BEAM_SWITCH_PENDING',
            0x100: 'GNSS_VALID',
            0x200: 'GNSS_REQUIRED',
            0x400: 'GNSS_PENDING',
        }),
        ('gnss_state_timer', 'uint'),
        ('reserved21', 'uint'),
        ('satellite_control_state', ControlState),
        ('beam_search_state', BeamState),
    )
)


EVENT_TRACES = (
    EVENT_TRACE_SATELLITE_GENERAL,
)


class GeoBeam(NimoIntEnum):
    """Geographic Beam identifiers mapped to readable names."""
    GLOBAL_BB = 0
    AMER_RB1 = 1
    AMER_RB2 = 2
    AMER_RB3 = 3
    AMER_RB4 = 4
    AMER_RB5 = 5
    AMER_RB6 = 6
    AMER_RB7 = 7
    AMER_RB8 = 8
    AMER_RB9 = 9
    AMER_RB10 = 10
    AMER_RB11 = 11
    AMER_RB12 = 12
    AMER_RB13 = 13
    AMER_RB14 = 14
    AMER_RB15 = 15
    AMER_RB16 = 16
    AMER_RB17 = 17
    AMER_RB18 = 18
    AMER_RB19 = 19
    AORW_SC = 61
    EMEA_RB1 = 21
    EMEA_RB2 = 22
    EMEA_RB3 = 23
    EMEA_RB4 = 24
    EMEA_RB5 = 25
    EMEA_RB6 = 26
    EMEA_RB7 = 27
    EMEA_RB8 = 28
    EMEA_RB9 = 29
    EMEA_RB10 = 30
    EMEA_RB11 = 31
    EMEA_RB12 = 32
    EMEA_RB13 = 33
    EMEA_RB14 = 34
    EMEA_RB15 = 35
    EMEA_RB16 = 36
    EMEA_RB17 = 37
    EMEA_RB18 = 38
    EMEA_RB19 = 39
    APAC_RB1 = 41
    APAC_RB2 = 42
    APAC_RB3 = 43
    APAC_RB4 = 44
    APAC_RB5 = 45
    APAC_RB6 = 46
    APAC_RB7 = 47
    APAC_RB8 = 48
    APAC_RB9 = 49
    APAC_RB10 = 50
    APAC_RB11 = 51
    APAC_RB12 = 52
    APAC_RB13 = 53
    APAC_RB14 = 54
    APAC_RB15 = 55
    APAC_RB16 = 56
    APAC_RB17 = 57
    APAC_RB18 = 58
    APAC_RB19 = 59
    # MEAS_RB10 = 90   # replaced by IOE
    # MEAS_RB11 = 91   # replaced by IOE
    # MEAS_RB12 = 92   # replaced by IOE
    # MEAS_RB15 = 93   # replaced by IOE
    IOE_RB1 = 101
    IOE_RB2 = 102
    IOE_RB3 = 103
    IOE_RB4 = 104
    IOE_RB5 = 105
    IOE_RB6 = 106
    IOE_RB7 = 107
    IOE_RB8 = 108
    IOE_RB9 = 109
    IOE_RB10 = 110
    IOE_RB11 = 111
    IOE_RB12 = 112
    IOE_RB13 = 113
    IOE_RB14 = 114
    IOE_RB15 = 115
    IOE_RB16 = 116
    IOE_RB17 = 117
    IOE_RB18 = 118
    IOE_RB19 = 119

    @property
    def satellite(self):
        return self.name.split('_')[0]

    @property
    def beam(self):
        return self.name.split('_')[1]
    
    @property
    def id(self):
        return self.value


class GeoSatellite(NimoFloatEnum):
    """Maps the Viasat/Inmarsat geostationary longitude supporting NIMO."""
    AMER = -98.0   # Inmarsat 4F3
    AORWSC = -54.0   # Inmarsat 3F5
    EMEA = 24.9   # Inmarsat 4AF4 aka Alphasat XL
    IOE = 63.5   # Inmarsat 6F1 previously IOR 3F1, MEAS 4F2
    APAC = 143.5   # Inmarsat 4F2 previously 4F1



"""Class for a Non-IP Modem using Orbcomm network protocols.

This module encapsulates specific AT commands from the following supported
NIMO modem manufacturers:

* **ORBCOMM**
* **Quectel**

Key concepts include modem properties, configuration, and satellite acquisition
details including signal level.

"""
import base64
import logging
import time
from dataclasses import dataclass
from enum import IntEnum
from threading import Event

from serial import Serial

from .atcommandbuffer import DEFAULT_AT_TIMEOUT, AtCommandBuffer
from .constants import (
    BAUDRATES,
    MSG_MO_MAX_SIZE,
    MSG_MO_NAME_MAX_LEN,
    MSG_MO_NAME_QMAX_LEN,
    AtErrorCode,
    BeamState,
    ControlState,
    DataFormat,
    GeoBeam,
    GnssMode,
    GnssModeOrbcomm,
    GnssModeQuectel,
    MessagePriority,
    MessageState,
    NetworkStatus,
    PowerMode,
    SignalLevelRegional,
    SignalQuality,
    UrcCode,
    WakeupPeriod,
    WakeupWay,
    WorkMode,
)
from .location import (
    ModemLocation,
    SatelliteLocation,
    get_location_from_nmea_data,
    get_satellite_location,
)
from .message import MoMessage, MtMessage, NimoMessage
from .nimoutils import iso_to_ts, vlog

VLOG_TAG = 'nimomodem'

_log = logging.getLogger(__name__)


class Manufacturer(IntEnum):
    """Supported NIMO modem implementations."""
    NONE = 0
    ORBCOMM = 1
    QUECTEL = 2


@dataclass
class AcquisitionInfo:
    """Details about the satellite acquisition state of the modem.
    
    Attributes:
        ctrl_state (ControlState): Primary network acquisition state.
        beam_state (BeamState): Secondary beam acquistion state.
        rssi (float): Signal indicator Carrier to Noise ratio (dB-Hz).
        vcid (int): Virtual carrier identifier for low-level sanity check.
    
    """
    ctrl_state: ControlState = ControlState.STOPPED
    beam_state: BeamState = BeamState.IDLE
    rssi: float = 0.0
    vcid: int = 0


class ModemError(Exception):
    """Base class for NIMO modem errors."""


class ModemTimeout(ModemError):
    """Serial response timeout."""


class ModemCrcConfig(ModemError):
    """Request/response mismatch of CRC presence."""


class ModemAtError(ModemError):
    """AT command related errors with error_code property."""
    @property
    def error_code(self) -> AtErrorCode:
        return AtErrorCode[self.args[0]]


class NimoModem:
    """A class for NIMO satellite IoT modem interaction."""
    # __slots__ = ('_modem', '_mobile_id',
    #              '_mfr_code', '_modem_booted', '_ready',
    #              )
    
    def __init__(self, serial_port: str, **kwargs) -> None:
        """Instantiate the NimoModem object.
        
        For additional kwargs see PySerial API:
        https://pyserial.readthedocs.io/en/latest/pyserial_api.html
        
        Args:
            serial_port (str): The path to the modem's serial port.
        
        Keyword Args:
            baudrate (int): The baud rate of the modem (default 9600)
        
        Raises:
            `ConnectionError` if unable to connect to the serial port.
        
        """
        try:
            self._serial = Serial(serial_port, **kwargs)
        except Exception as exc:
            raise ConnectionError('Unable to connect to serial') from exc
        self._modem: AtCommandBuffer = AtCommandBuffer(self._serial)
        self._baudrate: int = self._serial.baudrate
        self._is_connected: bool = False
        self._modem_booted: bool = False
        self._mobile_id: str = ''
        self._manufacturer: Manufacturer = Manufacturer.NONE
        self._ready = Event()
        self._ready.set()
    
    @property
    def is_ready(self) -> bool:
        return self._ready.is_set()
    
    @property
    def crc_enabled(self) -> bool:
        return self._modem.crc
    
    @property
    def modem_booted(self) -> bool:
        return self._modem_booted
    
    @property
    def _mfr(self) -> Manufacturer:
        """Used internally to support different manufacturer commands."""
        if not self._manufacturer:
            self.get_manufacturer()
        return self._manufacturer
    
    @property
    def _mo_msg_name_len_max(self) -> int:
        """Used internally to restrict the length of the MO message name."""
        maxlen = MSG_MO_NAME_MAX_LEN
        if self._mfr == Manufacturer.QUECTEL:
            maxlen = MSG_MO_NAME_QMAX_LEN
        return maxlen
    
    def _at_command_response(self,
                             command: str,
                             prefix: str = '',
                             timeout: int = DEFAULT_AT_TIMEOUT) -> str:
        """Send a command and return the response.
        
        Blocks until response has been received.
        
        Args:
            command (str): The AT command to send.
            prefix (str): Optional prefix to remove from response.
            timeout (int): Maximum time in seconds to wait for response.
        
        Raises:
            `ModemTimeout` if no response is received.
            `ModemCrcConfig` if CRC is not used on both request/response.
            `ModemAtError` for other cases of errored response.
        
        """
        self._ready.wait()
        self._ready.clear()
        try:
            self._modem.send_at_command(command)
            err = self._modem.read_at_response(prefix, timeout)
            if err == AtErrorCode.OK:
                return self._modem.get_response()
            elif err == AtErrorCode.TIMEOUT:
                raise ModemTimeout
            elif err == AtErrorCode.CRC_CONFIG_MISMATCH:
                raise ModemCrcConfig
            else:
                err = self.get_last_error_code()
                raise ModemAtError(err.name)
        finally:
            self._ready.set()
    
    def connect(self) -> None:
        """Attach to the modem via serial communications.
        
        Provided for backward compatibility. Connection is usually automatic
        when instantiating `NimoModem`.
        
        """
        if not self._serial.is_open:
            self._serial.open()
    
    def disconnect(self) -> None:
        """Detach from the modem serial communications.
        
        Provided for backward compatibility. Not typically required.
        
        """
        self._is_connected = False
        self._modem_booted = False
        if self._serial.is_open:
            self._serial.close()
    
    def is_connected(self) -> bool:
        """Indicates if the modem is responding to a basic AT query."""
        try:
            self._at_command_response('AT')
            self._is_connected = True
            self._modem_booted = True
            return True
        except ModemError:
            self._is_connected = False
            self._modem_booted = False
            return False
    
    @property
    def baudrate(self) -> int:
        """The baudrate of the serial connection."""
        return self._modem.serial.baudrate
    
    @baudrate.setter
    def baudrate(self, baudrate: int):
        """Set the baud rate of the modem and adjust the serial rate."""
        if baudrate not in [9600, 115200]:
            raise ValueError('Invalid baudrate')
        self._at_command_response(f'AT+IPR={baudrate}')
        self._modem.serial.baudrate = baudrate
    
    def retry_baudrate(self) -> bool:
        """"""
        for baud in BAUDRATES:
            self._modem.serial.baudrate = baud
            if self.is_connected():
                return True
        return False
    
    def await_boot(self, boot_timeout: int = 10) -> bool:
        """Indicates if a boot string is received within a timeout window.
        
        Use `is_connected` before waiting for boot.
        
        Args:
            boot_timeout (int): The maximum time to wait in seconds.
        
        Returns:
            True if a valid boot string was received inside the timeout.
        
        """
        boot_strings = ['ST Version', 'RDY']
        _log.debug('Awaiting modem boot string for %d seconds...', boot_timeout)
        rx_data = ''
        started = time.time()
        while time.time() - started < boot_timeout and not self._modem_booted:
            while self._modem.is_data_waiting():
                rx_data += self._modem.read_rx_buffer()
            if rx_data and any(b in rx_data for b in boot_strings):
                self._modem_booted = True
                _log.debug('Found boot string - clearing Rx buffer')
                while self._modem.is_data_waiting():
                    rx_data += self._modem.read_rx_buffer()
                break
        return self._modem_booted
    
    def get_last_error_code(self) -> AtErrorCode:
        """Get the last error code from the modem."""
        return AtErrorCode(int(self._at_command_response('ATS80?')))
    
    def initialize(self,
                   echo: bool = True,
                   verbose: bool = True,
                   ) -> bool:
        """Initialize the modem AT configuration for Echo and Verbose."""
        at_command = (f'ATZ;E{int(echo)};V{int(verbose)}')
        try:
            self._at_command_response(at_command)
            return True
        except ModemAtError as exc:
            if exc.error_code == AtErrorCode.CRC_CONFIG_MISMATCH:
                _log.info('Attempting re-initialize with CRC enabled')
                self._at_command_response(at_command)
                return True
            raise
    
    def set_crc(self, enable: bool = False) -> bool:
        """Enable or disable CRC error checking on the modem serial port."""
        try:
            self._at_command_response(f'AT%CRC={int(enable)}')
            return True
        except ModemCrcConfig:
            if ((self._modem.crc and enable) or
                (not self._modem.crc and not enable)):
                return True
            return False
    
    def reset_factory_config(self) -> None:
        """Reset the modem's factory default configuration."""
        self._at_command_response('AT&F')
    
    def save_config(self) -> None:
        """Store the current configuration to modem non-volatile memory."""
        self._at_command_response('AT&W')
    
    def get_mobile_id(self) -> str:
        """Get the modem's globally unique identifier."""
        if not self._mobile_id:
            try:
                self._mobile_id = self._at_command_response('AT+GSN', '+GSN:')
                if vlog(VLOG_TAG):
                    _log.debug('Cached Mobile ID %s', self._mobile_id)
            except ModemError:
                self._mobile_id = ''
                raise
        return self._mobile_id
    
    @property
    def _is_simulator(self) -> bool:
        return self.get_mobile_id().startswith('00000000')
    
    def get_manufacturer(self) -> str:
        """Get the manufacturer name."""
        if not self._manufacturer:
            try:
                mfr = self._at_command_response('ATI')
                if 'quectel' in mfr.lower():
                    self._manufacturer = Manufacturer.QUECTEL
                else:
                    if not any(m in mfr.lower()
                               for m in ['orbcomm', 'skywave']):
                        _log.warning('Unsupported manufacturer %s', mfr)
                    self._manufacturer = Manufacturer.ORBCOMM
                if vlog(VLOG_TAG):
                    _log.debug('Caching manufacturer: %s',
                               self._manufacturer.name)
            except ModemError:
                self._manufacturer = Manufacturer.NONE
                raise
        return self._manufacturer.name
    
    def get_model(self) -> str:
        """Get the manufacturer model name."""
        cmd = 'ATI4'
        if self._mfr == Manufacturer.QUECTEL:
            cmd = 'ATI'
        try:
            response = self._at_command_response(cmd)
            if response:
                if self._mfr == Manufacturer.QUECTEL:
                    response = response.split('\n')[1]
            return response
        except ModemError:
            return ''
    
    def get_firmware_version(self) -> str:
        """Get the modem's firmware version."""
        # TODO: Firmware structure with hardware, firmware, software?
        return self._at_command_response('AT+GMR', '+GMR:')
    
    def get_system_time(self) -> int:
        """Get the system/GNSS time from the modem."""
        try:
            nimo_time = self._at_command_response('AT%UTC', '%UTC:')
            iso_time = nimo_time.replace(' ', 'T') + 'Z'
            return iso_to_ts(iso_time)
        except ModemError:
            return 0
    
    def get_temperature(self) -> 'int|None':
        """Get the processor temperature in Celsius."""
        return int(int(self._at_command_response('ATS85?')) / 10)
    
    def is_transmit_allowed(self) -> bool:
        """Indicates if the modem is able to transmit data."""
        return self.get_network_status() == 5
    
    def is_blocked(self) -> bool:
        """Indicates if line-of-sight to the satellite is blocked."""
        return self.get_network_status() == 8
    
    def is_muted(self) -> bool:
        """Indicates if the modem has been muted (disallowed to transmit data).
        """
        return self.get_network_status() == 7
    
    def is_updating_network(self) -> bool:
        """Indicates if the modem is updating network information.
        
        The modem should not be powered down during a network update.
        
        """
        return self.get_network_status() == 4
    
    def get_network_status(self) -> NetworkStatus:
        """Get the current satellite acquisition status."""
        cmd = 'ATS54?'
        prefix = ''
        if self._mfr == Manufacturer.QUECTEL:
            cmd = 'AT+QREG?'
            prefix = '+QREG:'
        return NetworkStatus(int(self._at_command_response(cmd, prefix)))
    
    def get_rssi(self) -> float:
        """Get the current Received Signal Strength Indicator.
        
        Also referred to as SNR or C/N0 (dB-Hz)
        
        """
        cmd = 'ATS90=3 S91=1 S92=1 S116?'
        prefix = ''
        if self._mfr == Manufacturer.QUECTEL:
            cmd = 'AT+QSCN'
            prefix = '+QSCN:'
        return int(self._at_command_response(cmd, prefix)) / 100
    
    def get_signal_quality(self) -> SignalQuality:
        """Get a qualitative indicator from 0..5 of the satellite signal."""
        snr = self.get_rssi()
        if snr >= SignalLevelRegional.INVALID.value:
            return SignalQuality.WARNING
        if snr >= SignalLevelRegional.BARS_5.value:
            return SignalQuality.STRONG
        if snr >= SignalLevelRegional.BARS_4.value:
            return SignalQuality.GOOD
        if snr >= SignalLevelRegional.BARS_3.value:
            return SignalQuality.MID
        if snr >= SignalLevelRegional.BARS_2.value:
            return SignalQuality.LOW
        if snr >= SignalLevelRegional.BARS_1.value:
            return SignalQuality.WEAK
        return SignalQuality.NONE
    
    def get_acquisition_detail(self) -> AcquisitionInfo:
        """Get the detailed satellite acquisition status.
        
        Includes `acquisition_state`, `beamsearch_state`, `vcid` and `snr`
        indicators.
        
        """
        cmd = 'ATS90=3 S91=1 S92=1 S122? S123? S116? S101?'
        prefix = ''
        if self._mfr == Manufacturer.QUECTEL:
            cmd = 'AT+QEVNT=3,1'
            prefix = '+QEVNT:'
        result_str = self._at_command_response(cmd, prefix, timeout=10)
        if self._mfr == Manufacturer.ORBCOMM:
            results = [int(x) for x in result_str.split('\n')]
            ctrl_state = ControlState(results[0])
            beam_state = BeamState(results[1])
            rssi = float(results[2]) / 100
            vcid = results[3]
        elif self._mfr == Manufacturer.QUECTEL:
            # Workaround Quectel 20230731 documentation error says +QEVNT:
            result_str = result_str.replace('+QEVENT:', '').strip()
            results = [int(x) for x in result_str.split(',')]
            # <dataCount>,<signedBitmask>,<MTID>,<timestamp>,
            #   <class>,<subclass>,<priority>,<data0>,...
            data0 = 7   # list index where trace data starts
            ctrl_state = ControlState(results[data0+22])
            beam_state = BeamState(results[data0+23])
            rssi = float(results[data0+16]) / 100
            vcid = results[data0+1]
        return AcquisitionInfo(ctrl_state, beam_state, rssi, vcid)
    
    def send_data(self, data: bytes, **kwargs) -> 'str|MoMessage':
        """Submits data to send as a mobile-originated message.
        
        If a `message_name` is not supplied one will be generated using the
        least significant 8 digits of unix timestamp.
        
        Args:
            data (bytes): The data to send.
        
        Keyword Args:
            message_name (str): Optional handle for message in Tx queue. Max 8
                characters for Orbcomm modem or 12 for Quectel.
            priority (int): Optional priority 1 (highest) .. 4 (low, default).
                May use `MessagePriority`.
            codec_sin (int): Optional first byte of payload to add as a codec
                service identifier, must be in range 16..255.
            codec_min (int): Optional second byte of payload to add as a codec
                message identifier, must be in range 0..255.
            return_message (bool): If set, returns a `MoMessage` instead of the
                message handle.
        
        Returns:
            Message handle (str) or `MoMessage` if `return_message` kwarg is set.
        
        Raises:
            `ValueError` for various parameter limit violations.
        
        """
        data_size = len(data)
        msg_payload_sin_min = b''
        message_name = kwargs.get('message_name', '')
        priority = MessagePriority(kwargs.get('priority',
                                              MessagePriority.LOW.value))
        codec_sin = kwargs.get('codec_sin', -1)
        codec_min = kwargs.get('codec_min', -1)
        if codec_sin > -1:
            data_size += 1
            msg_payload_sin_min += codec_sin.to_bytes(1, 'big')
        if codec_min > -1:
            data_size += 1
            msg_payload_sin_min += codec_min.to_bytes(1, 'big')
        if not 2 <= data_size <= MSG_MO_MAX_SIZE:
            raise ValueError('Invalid mobile-originated message size')
        if message_name and len(message_name) > self._mo_msg_name_len_max:
            raise ValueError('Message name too long')
        data_index = 0
        if codec_sin <= -1:
            codec_sin = data[0]
            data_index += 1
            data_size -= 1
        if codec_sin not in range(16, 256):
            raise ValueError('Illegal first payload byte SIN must be 16..255')
        if codec_min <= -1:
            codec_min = data[1]
            data_index += 1
            data_size -= 1
        if codec_min > 255:
            raise ValueError('Invalid second payload byte MIN must be 0..255')
        max_name_len = self._mo_msg_name_len_max
        if message_name and len(message_name) > max_name_len:
            raise ValueError(f'Invalid message name longer than {max_name_len}')
        if len(message_name) == 0:
            message_name = f'{int(time.time())}'[-max_name_len:]
        # Convert to base64 string for serial efficiency
        #   no effect on OTA size, modem always decodes and sends raw bytes OTA
        data_format = DataFormat.BASE64
        formatted_data = base64.b64encode(data[2:]).decode('utf-8')
        cmd = 'AT%MGRT='
        codec_sep = '.'
        if self._mfr == Manufacturer.QUECTEL:
            cmd = 'AT+QSMGT='
            codec_sep = ','
        cmd = (f'{cmd}"{message_name}",{priority},{codec_sin}{codec_sep}'
               f'{codec_min},{data_format},{formatted_data}')
        self._at_command_response(cmd)
        if kwargs.get('return_message', False) is True:
            return MoMessage(message_name, priority, MessageState.TX_READY,
                                payload=(msg_payload_sin_min + data))
        return message_name
    
    def send_text(self, text: str, **kwargs) -> 'str|MoMessage':
        """Submits a text string to send as data.
        
        If `codec_sin` kwarg is not provided 128 is prepended as the first byte.
        If `codec_min` kwarg is not provided 1 is prepended as the second byte.
        Other kwargs as per `send_data`.
        
        Args:
            text (str): The text message to send.
        
        Returns:
            (str) The message name assigned or MoMessage if kwarg
                `return_message` is set.
        
        """
        data = b''
        codec_sin = int(kwargs.get('codec_sin', 128))
        data += codec_sin.to_bytes(1, 'big')
        codec_min = int(kwargs.get('codec_min', 1))
        data += codec_min.to_bytes(1, 'big')
        data += text.encode()
        flowthru = ['message_name', 'priority', 'return_message']
        next_kwargs = { k:v for k, v in kwargs if k in flowthru }
        return self.send_data(data, **next_kwargs)
    
    def cancel_mo_message(self, message_name: str) -> bool:
        """Attempts to cancel a previously submitted mobile-originated message.
        
        Args:
            message_name (str): The mobile-originated message handle to delete.
        
        """
        _log.debug('Attempting to cancel MO message %s', message_name)
        cmd = 'AT%MGRC'
        if self._mfr == Manufacturer.QUECTEL:
            cmd = 'AT+QSMGC'
        cmd += f'="{message_name}"'
        self._at_command_response(cmd)
        message_states = self.get_mo_message_states(message_name)
        if len(message_states) > 0:
            state = message_states[0].state
            if state == MessageState.TX_CANCELLED:
                return True
        elif self._is_simulator:
            return True
        _log.warn('Failed to cancel message %s', message_name)
        return False
    
    def get_mo_message_states(self, message_name: str = '') -> 'list[MoMessage]':
        """Get a list of mobile-originated message states in the modem Tx queue.
        
        Args:
            message_name (str): Optional filter on message name.
        
        Returns:
            A list of `MoMessage` objects including state and metadata.
        
        """
        cmd = 'AT%MGRS'
        prefix = '%MGRS:'
        if self._mfr == Manufacturer.QUECTEL:
            cmd = 'AT+QSMGS'
            prefix = '+QSMGS:'
        if message_name and not self._is_simulator:
            # Orbcomm Modem Simulator returns ERROR for %MGRS= command
            cmd += f'="{message_name}"'
        response_str = self._at_command_response(cmd, prefix)
        return self._parse_message_states(response_str, is_mo=True)
    
    def _parse_message_states(self,
                              response_str: str,
                              is_mo: bool,
                              ) -> 'list[NimoMessage]':
        """Parses textual metadata to build a SatelliteMessageState."""
        mo_states = []
        if not response_str:
            return mo_states
        if vlog(VLOG_TAG):
            _log.debug('Parsing %s message states from %s',
                       'MO' if is_mo else 'MT', response_str)
        states_meta = [m for m in response_str.split('\n') if m != '']
        for meta in states_meta:
            message = MoMessage() if is_mo else MtMessage()
            for field_idx, field_data in enumerate(meta.split(',')):
                self._update_message_state(message, field_idx,
                                           field_data, is_mo)
            mo_states.append(message)
        return mo_states
    
    def _update_message_state(self,
                              message_state: NimoMessage,
                              field_idx: int,
                              field_data: str,
                              is_mo: bool) -> None:
        """Parse textual metadata to update a message's state."""
        if vlog(VLOG_TAG):
            _log.debug('Parsing %s message state index %d: %s',
                       'MO' if is_mo else 'MT', field_idx, field_data)
        mfr = self._mfr
        if field_idx == 0:
            message_state.name = field_data.replace('"', '')
            if vlog(VLOG_TAG):
                _log.debug('Message name: %s', message_state.name)
        elif field_idx == 1 and mfr == Manufacturer.ORBCOMM:
            if vlog(VLOG_TAG):
                _log.debug('Ignoring msgNum %s', field_data)
        elif ((field_idx == 2 and mfr == Manufacturer.ORBCOMM) or
              (field_idx == 1 and mfr == Manufacturer.QUECTEL)):
            message_state.priority = MessagePriority(int(field_data))
            if vlog(VLOG_TAG):
                _log.debug('Message priority %s', message_state.priority.name)
        elif ((field_idx == 3 and mfr == Manufacturer.ORBCOMM) or
              (field_idx == 2 and mfr == Manufacturer.QUECTEL)):
            if vlog(VLOG_TAG):
                _log.debug('Ignoring codec SIN %s', field_data)
        elif ((field_idx == 4 and mfr == Manufacturer.ORBCOMM) or
              (field_idx == 3 and mfr == Manufacturer.QUECTEL)):
            message_state.state = MessageState(int(field_data))
            _log.debug('Message state: %s', message_state.state.name)
        elif ((field_idx == 5 and mfr == Manufacturer.ORBCOMM) or
              (field_idx == 4 and mfr == Manufacturer.QUECTEL)):
            message_state.length = int(field_data)
            _log.debug('Message size: %d bytes', message_state.length)
        elif ((field_idx == 6 and mfr == Manufacturer.ORBCOMM) or
              (field_idx == 5 and mfr == Manufacturer.QUECTEL)):
            message_state.bytes_delivered = int(field_data)
            _log.debug('Bytes delivered: %d', message_state.bytes_delivered)
        else:
            _log.warning('Unhandled field index %d (%s) for manufacturer %s',
                         field_idx, 'MO' if is_mo else 'MT', mfr.name)
    
    def get_mt_message_states(self, message_name: str = '') -> 'list[MtMessage]':
        """Get a list of mobile-terminated message states in the modem Tx queue.
        
        Args:
            message_name (str): Optional filter on message name.
        
        Returns:
            A list of `MtMessage` objects including state and metadata.
        
        """
        cmd = 'AT%MGFN' if not message_name else 'AT%MGFS'
        prefix = '%MGFN:' if not message_name else 'AT%MGFS:'
        if self._mfr == Manufacturer.QUECTEL:
            cmd = 'AT+QRMGN' if not message_name else 'AT+QRMGS'
            prefix = '+QRMGN:' if not message_name else '+QRMGS:'
        if message_name and not self._is_simulator:
            cmd += f'="{message_name}"'
        response_str = self._at_command_response(cmd, prefix)
        return self._parse_message_states(response_str, is_mo=False)
    
    def get_mt_message(self, message_name: str) -> 'MtMessage|None':
        """Get a mobile-terminated message from the modem's Rx queue by name."""
        cmd = 'AT%MGFG'
        prefix = '%MGFG:'
        if self._mfr == Manufacturer.QUECTEL:
            cmd = 'AT+GRMGR'
            prefix = '+GRMGR:'
        data_format = DataFormat.BASE64
        cmd += f'="{message_name}",{data_format}'
        response = self._at_command_response(cmd, prefix)
        if response:
            return self._parse_mt_message(response)
        return None
    
    def _parse_mt_message(self, meta: str) -> MtMessage:
        """Parse textual metadata to build a MtMessage."""
        if vlog(VLOG_TAG):
            _log.debug('Parsing MT message from meta: %s', meta)
        data_includes_sin = False
        mfr = self._mfr
        message = MtMessage()
        for field_idx, field_data in enumerate(meta.split(',')):
            if field_idx == 0:
                message.name = field_data.replace('"', '')
                if vlog(VLOG_TAG):
                    _log.debug('Message name: %s', message.name)
            elif (field_idx == 1 and mfr == Manufacturer.ORBCOMM):
                if vlog(VLOG_TAG):
                    _log.debug('Ignoring msgNum %s', field_data)
            elif (field_idx == 2 and mfr == Manufacturer.ORBCOMM):
                message.priority = MessagePriority(int(field_data))
                if vlog(VLOG_TAG):
                    _log.debug('Message priority %s', message.priority.name)
            elif ((field_idx == 3 and mfr == Manufacturer.ORBCOMM) or
                  (field_idx == 1 and mfr == Manufacturer.QUECTEL)):
                codec_sin = int(field_data)
                if not data_includes_sin:
                    message.payload += codec_sin.to_bytes(1, 'big')
                if vlog(VLOG_TAG):
                    _log.debug('Added SIN as first payload byte: %d', codec_sin)
            elif (field_idx == 4 and mfr == Manufacturer.ORBCOMM):
                message.state = MessageState(int(field_data))
                if vlog(VLOG_TAG):
                    _log.debug('Message state %s', message.state.name)
            elif ((field_idx == 5 and mfr == Manufacturer.ORBCOMM) or
                  (field_idx == 2 and mfr == Manufacturer.QUECTEL)):
                message.length = int(field_data)
                if vlog(VLOG_TAG):
                    _log.debug('Message size: %d bytes', message.length)
            elif ((field_idx == 6 and mfr == Manufacturer.ORBCOMM) or
                  (field_idx == 3 and mfr == Manufacturer.QUECTEL)):
                data_format = DataFormat(int(field_data))
                if vlog(VLOG_TAG):
                    _log.debug('Data format %s', data_format.name)
            elif ((field_idx == 7 and mfr == Manufacturer.ORBCOMM) or
                  (field_idx == 4 and mfr == Manufacturer.QUECTEL)):
                if vlog(VLOG_TAG):
                    _log.debug('Decoding payload from: %s', field_data)
                if message.length > 0:
                    if data_format == DataFormat.BASE64:
                        message.payload += base64.b64decode(field_data)
                    elif data_format == DataFormat.HEX:
                        message.payload += bytes.fromhex(field_data)
                    else:   # DataFormat.TEXT
                        message.payload += field_data.encode()
                    if message.length != len(message.payload):
                        _log.warn('Message length mismatch')
        return message
    
    def delete_mt_message(self, message_name: str) -> bool:
        """Remove a mobile-terminated message from the modem's Rx queue."""
        cmd = 'AT%MGFM'
        if self._mfr == Manufacturer.QUECTEL:
            cmd = 'AT+QRMGM'
        cmd += f'="{message_name}"'
        self._at_command_response(cmd)
        check = self.get_mt_message_states(message_name)
        if check and check[0].state == MessageState.RX_RETRIEVED:
            return True
        return False
    
    def receive_data(self, message_name: str) -> 'bytes|None':
        """Get the raw data from a mobile-terminated message."""
        message = self.get_mt_message(message_name)
        if message:
            return message.payload
        return None
    
    def get_gnss_mode(self) -> GnssMode:
        """Get the modem's GNSS receiver mode."""
        cmd = 'ATS39?'
        prefix = ''
        if self._mfr == Manufacturer.QUECTEL:
            cmd = 'AT+QGNSSMOD?'
            prefix = '+QGNSSMOD:'
        response = self._at_command_response(cmd, prefix)
        if self._mfr == Manufacturer.QUECTEL:
            return GnssModeQuectel(int(response))
        return GnssModeOrbcomm(int(response))
    
    def set_gnss_mode(self, gnss_mode: GnssMode) -> None:
        """Get the modem's GNSS receiver mode."""
        cmd = f'ATS39={gnss_mode}'
        prefix = ''
        if self._mfr == Manufacturer.QUECTEL:
            if not GnssModeQuectel.is_valid(gnss_mode):
                raise ValueError('Invalid GNSS mode')
            cmd = f'AT+QGNSSMOD={gnss_mode}'
            prefix = '+QGNSSMOD:'
        else:
            if not GnssModeOrbcomm.is_valid(gnss_mode):
                raise ValueError('Invalid GNSS mode')
        self._at_command_response(cmd, prefix)
    
    def get_gnss_continuous(self) -> int:
        """Get the modem's GNSS continuous refresh interval in seconds."""
        cmd = 'ATS55?'
        prefix = ''
        if self._mfr == Manufacturer.QUECTEL:
            cmd = 'AT+QGNSSCW?'
            prefix = '+QGNSSCW:'
        return int(self._at_command_response(cmd, prefix))
    
    def set_gnss_continuous(self, interval: int) -> None:
        """Set the modem's GNSS continuous refresh interval in seconds.
        
        Args:
            interval (int): Automatic update interval 0..30 seconds.
        
        Returns:
            `True` if successful.
        
        Raises:
            `ValueError` if invalid interval is specified.
        
        """
        if interval not in range (0, 31):
            raise ValueError('Invalid GNSS refresh interval')
        cmd = f'ATS55={interval}'
        if self._mfr == Manufacturer.QUECTEL:
            cmd = f'AT+QGNSSCW={interval}'
        self._at_command_response(cmd)
    
    def get_nmea_data(self,
                      stale_secs: int = 1,
                      wait_secs: int = 35,
                      rmc: bool = True,
                      gga: bool = True,
                      gsa: bool = True,
                      gsv: bool = False,
                      ) -> str:
        """Get a set of NMEA data detailing the modem's location.
        
        Args:
            stale_secs (int): Maximum cached fix age to use in seconds.
            wait_secs (int): Maximum duration to wait for a fix in seconds.
            rmc (bool): Include Recommended Minimum data.
            gga (bool): Include altitude and fix quality data.
            gsa (bool): Include Dilution of Precision data.
            gsv (bool): Include verbose GNSS satellite details.
        
        """
        cmd = 'AT%GPS'
        prefix = '%GPS:'
        if self._mfr == Manufacturer.QUECTEL:
            cmd = 'AT+QGNSS'
            prefix = '+QGNSS:'
        cmd += f'={stale_secs},{wait_secs}'
        if rmc:
            cmd += ',"RMC"'
        if gga:
            cmd += ',"GGA"'
        if gsa:
            cmd += ',"GSA"'
        if gsv:
            cmd += ',"GSV"'
        try:
            response = self._at_command_response(cmd, prefix, wait_secs + 5)
            return response
        except ModemAtError as exc:
            if exc.error_code != AtErrorCode.GNSS_TIMEOUT:
                raise
        return ''
    
    def get_location(self,
                     stale_secs: int = 1,
                     wait_secs: int = 35) -> 'ModemLocation|None':
        """Get the modem's location.
        
        Args:
            stale_secs (int): Maximum cached fix age to use in seconds.
            wait_secs (int): Maximum duration to wait for a fix in seconds.
        
        Returns:
            ModemLocation object if GNSS does not time out waiting for fix.
        
        """
        nmea_data = self.get_nmea_data(stale_secs, wait_secs)
        if nmea_data:
            return get_location_from_nmea_data(nmea_data)
        return None
    
    def get_satellite_info(self) -> 'SatelliteLocation|None':
        """Get the satellite's information including azimuth and elevation.
        
        Derives which satellite/GeoBeam is used from trace class 3 subclass 5.
        
        Returns:
            `SatelliteLocation` object (azimuth, elevation) if determinable.
        
        """
        geobeam = None
        modem_location = self.get_location()
        if (modem_location is not None and
            self.get_network_status() > NetworkStatus.RX_SEARCHING):
            # satellite has been found
            cmd = 'ATS90=3 S91=5 S92=1 S102?'
            prefix = ''
            if self._mfr == Manufacturer.QUECTEL:
                cmd = 'AT+QEVNT=3,5'
                prefix = '+QEVNT:'
            response = self._at_command_response(cmd, prefix)
            if self._mfr == Manufacturer.QUECTEL:
                # workaround documentation error
                response = response.replace('+QEVENT:', '').strip()
                response = response.split(',')[9]
            geobeam = GeoBeam(int(response))
        satellite_loc = get_satellite_location(modem_location, geobeam)
        return satellite_loc if satellite_loc.azimuth else None
    
    def get_event_mask(self) -> int:
        """Get the set of monitored events that trigger event notification."""
        if self._mfr != Manufacturer.ORBCOMM:
            raise ModemError('Operation not supported by this modem')
        cmd = 'ATS88?'
        return int(self._at_command_response(cmd))
    
    def set_event_mask(self, event_mask: int) -> None:
        """Set monitored events that trigger event notification."""
        if self._mfr != Manufacturer.ORBCOMM:
            raise ModemError('Operation not supported by this modem')
        max_bits = 12
        if not isinstance(event_mask, int) or event_mask > 2**max_bits-1:
            raise ValueError('Invalid event bitmask')
        cmd = f'ATS88={event_mask}'
        self._at_command_response(cmd)
    
    def get_events_asserted_mask(self) -> int:
        """Get the set of events that are active following a notification."""
        if self._mfr != Manufacturer.ORBCOMM:
            raise ModemError('Operation not supported by this modem')
        cmd = 'ATS89?'
        return int(self._at_command_response(cmd))
    
    def get_trace_event_monitor(self,
                                asserted_only: bool = False,
                                ) -> 'list[tuple[int, int]]':
        """Get the list of monitored Trace Events.
        
        Args:
            asserted_only (bool): If True returns only asserted monitored events
        
        Returns:
            A list of tuples (trace_class, trace_subclass)
        
        Raises:
            `ModemError` if unsupported by the modem type.
        
        """
        if self._mfr != Manufacturer.ORBCOMM:
            raise ModemError('Operation not supported by this modem')
        cmd = 'AT%EVMON'
        prefix = '%EVMON:'
        trace_events = []
        events = self._at_command_response(cmd, prefix).split(',')
        for event in events:
            trace_class = int(event.split('.')[0])
            trace_subclass = int(event.split('.')[1].replace('*', ''))
            if not asserted_only or event.endswith('*'):
                trace_events.append((trace_class, trace_subclass))
        return trace_events
    
    def set_trace_event_monitor(self, events: 'list[tuple[int, int]]') -> None:
        """Set the list of monitored trace events."""
        cmd = 'AT%EVMON='
        for event in events:
            if not cmd.endswith('='):
                cmd += ','
            cmd += f'{event[0].event[1]}'
        self._at_command_response(cmd)
    
    def get_trace_events_cached(self) -> 'list[tuple[int, int]]':
        """Get a list of trace events cached."""
        return self.get_trace_event_monitor(True)
    
    def get_trace_event_data(self,
                             event: 'tuple[int, int]',
                             decode: bool = False,
                             ) -> 'list[int]|dict[str, int]':
        """Get the trace event data.
        
        Args:
            event (tuple): The trace (class, subclass)
            decode (bool): Decodes raw data to dictionary (not implemented)
        
        """
        cmd = f'AT%EVNT={event[0]},{event[1]}'
        prefix = '%EVNT:'
        if self._mfr == Manufacturer.QUECTEL:
            cmd = cmd.replace('%EVNT', '+QEVNT')
            prefix = '+QEVENT:'   # documented as +QEVNT
        trace = self._at_command_response(cmd, prefix)
        if decode:
            raise NotImplementedError
        return [int(i) for i in trace.split(',')]
            
    def get_urc_ctl(self) -> int:
        """Get the event list that trigger Unsolicited Report Codes."""
        if self._mfr != Manufacturer.QUECTEL:
            raise ValueError('Modem does not support this feature')
        cmd = 'AT+QURCCTL?'
        prefix = '+QURCCTL:'
        return int(self._at_command_response(cmd, prefix), 16)
    
    def set_urc_ctl(self, qurc_mask: int) -> None:
        """Set the event list that trigger Unsolicited Report Codes."""
        if self._mfr != Manufacturer.QUECTEL:
            raise ValueError('Modem does not support this feature')
        cmd = f'AT+QURCCTL=0x{qurc_mask:04X}'
        self._at_command_response(cmd)
    
    def get_urc(self) -> 'UrcCode|None':
        """Get the pending Unsolicited Result Code if one is present."""
        if self._mfr != Manufacturer.QUECTEL:
            raise ValueError('Modem does not support this feature')
        eol = '\r\n' if self._modem.verbose else '\r'
        result = self._modem.read_line(eol)
        if result:
            result = result.replace('+QURC:', '').strip()
            try:
                return UrcCode(int(result))
            except ValueError:
                return UrcCode[result]
        return None
    
    def get_power_mode(self) -> PowerMode:
        """Get the modem's power mode configuration."""
        cmd = 'ATS50?'
        prefix = ''
        if self._mfr == Manufacturer.QUECTEL:
            cmd = 'AT+QPMD?'
            prefix = '+QPMD:'
        return PowerMode(int(self._at_command_response(cmd, prefix)))
    
    def set_power_mode(self, power_mode: PowerMode) -> None:
        """Set the modem's power mode configuration."""
        if not PowerMode.is_valid(power_mode):
            raise ValueError('Invalid Power Mode')
        cmd = f'ATS50={power_mode}'
        if self._mfr == Manufacturer.QUECTEL:
            cmd = f'AT+QPMD={power_mode}'
        self._at_command_response(cmd)
    
    def get_wakeup_period(self) -> WakeupPeriod:
        """Get the modem's wakeup period configuration."""
        cmd = 'ATS51?'
        prefix = ''
        if self._mfr == Manufacturer.QUECTEL:
            cmd = 'AT+QWKUPCFG?'
            prefix = '+QWKUPCFG:'
        if self._mfr == Manufacturer.QUECTEL:
            return WakeupPeriod(int(
                self._at_command_response(cmd, prefix).split(',')[0]))
        return WakeupPeriod(int(self._at_command_response(cmd, prefix)))
    
    def set_wakeup_period(self,
                          wakeup_period: WakeupPeriod,
                          wakeup_way: 'WakeupWay|None' = None,
                          ) -> None:
        """Set the modem's wakeup period configuration.
        
        The configuration does not update until confimed by the network.
        
        """
        if not WakeupPeriod.is_valid(wakeup_period):
            raise ValueError('Invalid wakeup period')
        cmd = f'ATS51={wakeup_period}'
        if self._mfr == Manufacturer.QUECTEL:
            if wakeup_way is None:
                query = self._at_command_response('AT+QWKUPCFG?', '+QWKUPCFG:')
                wakeup_way = WakeupWay(int(query.split(',')[1]))
            cmd = f'AT+QWKUPCFG={wakeup_period},{wakeup_way}'
        self._at_command_response(cmd)
    
    def get_wakeup_way(self) -> WakeupWay:
        """Get the modem wakeup method."""
        if self._mfr != Manufacturer.QUECTEL:
            raise ModemError('Operation not supported by this modem')
        cmd = 'AT+QWKUPCFG?'
        prefix = '+QWKUPCFG:'
        wakeup_way = self._at_command_response(cmd, prefix).split(',')[1]
        return WakeupWay(int(wakeup_way))
    
    def power_down(self) -> None:
        """Prepare the modem for power-down."""
        cmd = 'AT%OFF'
        if self._mfr == Manufacturer.QUECTEL:
            cmd = 'AT+QPOWD=2'
        self._at_command_response(cmd)
    
    def get_workmode(self) -> WorkMode:
        """Get the modem working mode."""
        if self._mfr != Manufacturer.QUECTEL:
            raise ModemError('Operation not supported by this modem')
        cmd = 'AT+QMOD?'
        prefix = '+QMOD:'
        return WorkMode(int(self._at_command_response(cmd, prefix)))
    
    def set_workmode(self, workmode: WorkMode) -> None:
        """Set the modem working mode."""
        if self._mfr != Manufacturer.QUECTEL:
            raise ModemError('Operation not supported by this modem')
        if not WorkMode.is_valid(workmode):
            raise ValueError('Invalid workmode')
        cmd = f'AT+QMOD={workmode}'
        self._at_command_response(cmd)
    
    def get_deepsleep_enable(self) -> bool:
        """Get the deepsleep configuration flag."""
        if self._mfr != Manufacturer.QUECTEL:
            raise ModemError('Operation not supported by this modem')
        cmd = 'AT+QSCLK?'
        prefix = '+QSCLK:'
        return bool(int(self._at_command_response(cmd, prefix)))
    
    def set_deepsleep_enable(self, enable: bool) -> None:
        """Set the deepsleep configuration flag."""
        if self._mfr != Manufacturer.QUECTEL:
            raise ModemError('Operation not supported by this modem')
        self._at_command_response(f'AT+QSCLK={int(enable)}')
    
    def get_register(self, s_register_number: int) -> 'int|None':
        """Get a modem register value."""
        cmd = f'ATS{s_register_number}?'
        return int(self._at_command_response(cmd))
    
    def set_register(self, s_register_number: int, value: int) -> None:
        """Set a modem register value."""
        cmd = f'ATS{s_register_number}={value}'
        self._at_command_response(cmd)
    
    def get_all_registers(self) -> dict:
        """Get a dictionary of modem register values."""
        raise NotImplementedError



"""Class and methods for managing messages submitted/retrieved to a NIMO modem.

This module encapsulates the key concepts of operation for:

* **Mobile-Originated** (aka From-Terminal, Return) messages
* **Mobile-Terminated** (aka To-Terminal, Forward) messages
* Codec key concepts using **SIN** (Service Identification Number) and
**MIN** (Message Identification Number) as the first 2 bytes of message payload.
* Local message metadata such as **state** and **priority**.

"""
import logging

from .constants import (
    MSG_MO_NAME_MAX_LEN,
    MSG_MO_NAME_QMAX_LEN,
    MessagePriority,
    MessageState,
)
from .nimoutils import vlog

VLOG_TAG = 'nimomessage'

_log = logging.getLogger(__name__)


class NimoMessage:
    """A satellite message.
    
    Attributes:
        name (str): The message name in the modem's queue.
        priority (MessagePriority): The priority assigned.
        state (MessageState): The message state.
        payload (bytes): The data content of the message.
        codec_sin (int): The first byte of payload.
        codec_min (int): The second byte of payload.
        length (int): The size of the payload.
    
    """
    def __init__(self,
                 name: str = '',
                 priority: MessagePriority = MessagePriority.NONE,
                 state: MessageState = MessageState.UNAVAILABLE,
                 length: int = 0,
                 bytes_delivered: int = 0,
                 payload: bytes = b'',
                 ) -> None:
        self._message_name: str = ''
        if name:
            self.name = name
        self._priority = MessagePriority.NONE
        if priority is not None:
            self.priority = priority
        self._state = MessageState.UNAVAILABLE
        if state is not None:
            self.state = state
        self._length: int = 0
        if length:
            self.length = length
        self._bytes_delivered: int = 0
        if bytes_delivered:
            self.bytes_delivered = bytes_delivered
        self.payload: bytes = payload
    
    @property
    def name(self) -> str:
        return self._message_name
    
    @name.setter
    def name(self, message_name: str):
        if not isinstance(message_name, str) or len(message_name) == 0:
            raise ValueError('Invalid message name')
        self._message_name = message_name
    
    @property
    def priority(self) -> MessagePriority:
        return self._priority
    
    @priority.setter
    def priority(self, value: MessagePriority):
        if not MessagePriority.is_valid(value):
            raise ValueError('Invalid MessagePriority')
        self._priority = MessagePriority(value)
    
    @property
    def state(self) -> MessageState:
        return self._state
    
    @state.setter
    def state(self, value: MessageState):
        if not MessageState.is_valid(value):
            raise ValueError('Invalid MessagePriority')
        self._state = MessageState(value)
    
    @property
    def codec_sin(self) -> int:
        if self.payload and len(self.payload) > 2:
            return int(self.payload[0])
        return -1
        
    @property
    def codec_min(self) -> int:
        if self.payload and len(self.payload) > 2:
            return int(self.payload[1])
        return -1
    
    @property
    def length(self) -> int:
        if (self._length < len(self.payload) or
            len(self.payload) > 2 and self._length != len(self.payload)):
            # update to actual length
            if vlog(VLOG_TAG):
                _log.debug('Updating message length to align with payload')
            self._length = len(self.payload)
        return self._length
    
    @length.setter
    def length(self, value: int):
        if not isinstance(value, int) or value <= 0:
            raise ValueError('Invalid message length')
        if len(self.payload) > 1 and value != len(self.payload):
            # >1 condition allows for SIN peel-off from MT parsing
            _log.warn('Length mismatch with payload size')
        self._length = value
    
    @property
    def bytes_delivered(self) -> int:
        return self._bytes_delivered
    
    @bytes_delivered.setter
    def bytes_delivered(self, value: int):
        if not isinstance(value, int) or value < 0:
            raise ValueError('Invalid bytes delivered')
        if value > self.length:
            _log.error('Bytes delivered mismatch with message length')
            return
        self._bytes_delivered = value


class MoMessage(NimoMessage):
    """A Mobile-Originated Message."""
    @property
    def name(self) -> str:
        return self._message_name
    
    @name.setter
    def name(self, message_name: str):
        msg_mo_name_max_len = max(MSG_MO_NAME_MAX_LEN, MSG_MO_NAME_QMAX_LEN)
        if (not isinstance(message_name, str) or
            not 0 < len(message_name) <= msg_mo_name_max_len):
            raise ValueError('Invalid message name')
        self._message_name = message_name
    

class MtMessage(NimoMessage):
    """A Mobile-Terminated message."""
    @property
    def bytes_delivered(self) -> int:
        if self._bytes_delivered < self.length:
            # bytes delivered not updated during parsing - update
            self._bytes_delivered = self.length
        return self._bytes_delivered

    @bytes_delivered.setter
    def bytes_delivered(self, value: int):
        if not isinstance(value, int) or value < 0:
            raise ValueError('Invalid bytes delivered')
        if value > self.length:
            _log.error('Bytes delivered mismatch with message length')
            return
        self._bytes_delivered = value



"""Implementation of CCIT-16-CRC for use with NIMO modems.

This module enables CRC error checking on a serial AT command interface, useful
for increasing robustness in electrically noisy environments or long cable runs.

"""
import logging

from .nimoutils import dprint, vlog

POLYNOMIAL = 0x1021
CRCXMODEM_SEPARATOR = '*'
VLOG_TAG = 'crcxmodem'

_log = logging.getLogger(__name__)
_crcxmodem_table: 'list[int]' = []
_crc_table_is_initialized: bool = False


def apply_crc(at_command: str, sep: str = CRCXMODEM_SEPARATOR) -> str:
    """Applies a CRC-16-CCITT checksum to the at_command."""
    crc = _calculate_crc(at_command)
    hex_crc = f'{crc:04X}'
    if vlog(VLOG_TAG):
        _log.debug('Applying CRC: %d -> %s', crc, hex_crc)
    return at_command + sep + hex_crc


def validate_crc(response: str, sep = CRCXMODEM_SEPARATOR) -> bool:
    """Validates a modem response with checksum."""
    if sep not in response:
        _log.warning('No CRC in response %s', dprint(response))
        return False
    if vlog(VLOG_TAG):
        _log.debug('Validating CRC for %s', dprint(response))
    response_crc = response.split(sep, 1)[1]
    expected = _calculate_crc(response)
    received = int(response_crc, 16)
    return expected == received


def _initial_crc(c: int) -> int:
    """Generates an initial CRC for each element in the table."""
    crc = 0
    c = c << 8
    for i in range(8):
        if (crc ^ c) & 0x8000:
            crc = (crc << 1) ^ POLYNOMIAL
        else:
            crc = crc << 1
        c = c << 1
    return crc


def _initialize_crc_table() -> bool:
    """Initializes the table of CRC values for lookup."""
    global _crcxmodem_table
    global _crc_table_is_initialized
    if not _crc_table_is_initialized:
        if vlog(VLOG_TAG):
            _log.debug('Initializing CRC table')
        _crcxmodem_table = [_initial_crc(i) for i in range(256)]
        _crc_table_is_initialized = True
    return _crc_table_is_initialized


def _update_crc(crc: int, c: int) -> int:
    """Call iteratively to build a CRC a character at a time."""
    global _crcxmodem_table
    cc = 0xFF & c
    tmp = (crc >> 8) ^ cc
    crc = (crc << 8) ^ _crcxmodem_table[tmp & 0xFF]
    crc = crc & 0xFFFF
    return crc

def _calculate_crc(string: str,
                   initial_value: int = 0xFFFF,
                   sep: str = CRCXMODEM_SEPARATOR) -> int:
    """Calculates the CRC of a string."""
    _initialize_crc_table()
    crc: int = initial_value
    for c in string:
        if c == sep:
            break
        crc = _update_crc(crc, ord(c))
    return crc



"""AT command buffer parsing for a NIMO modem.

This module provides a bidirectional buffer model for sendind and receiving
AT commands on a serial link to a satellite modem following V.25 spec.

"""
import logging
import threading
import time

from serial import Serial

from .constants import AtErrorCode, AtParsingState
from .crcxmodem import apply_crc, validate_crc
from .nimoutils import dprint, vlog

VLOG_TAG = 'atcommand'
DEFAULT_AT_TIMEOUT = 3   # seconds
VRES_OK = '\r\nOK\r\n'
VRES_ERR = '\r\nERROR\r\n'
RES_OK = '0\r'
RES_ERR = '4\r'

_log = logging.getLogger(__name__)


class AtCommandBuffer:
    """A command/response buffer for communicating with a NIMO modem.
    
    Attributes:
        serial (serial.Serial): The Serial object/stream used for communications
        echo (bool): Echo enabled on commands (default True)
        verbose (bool): Verbose response codes (default True)
        quiet (bool): Suppressed response codes (default False)
        crc (bool): CRC/checksum request/response enabled (default False)
        ready (threading.Event): Blocks new requests until pending response
            has completed or timed out.
    
    """
    def __init__(self, serial: Serial) -> None:
        self.echo: bool = True
        self.verbose: bool = True
        self.quiet: bool = False
        self.crc: bool = False
        if not isinstance(serial, Serial):
            raise ValueError('Invalid serial port')
        if vlog(VLOG_TAG):
            _log.debug('Connecting to %s (%d baud)',
                       serial.name, serial.baudrate)
        self.serial = serial
        self._char_delay: float = 8 / serial.baudrate
        self._pending_command: str = None
        self._rx_buffer: str = ''
        self.ready = threading.Event()
        self.ready.set()
    
    def is_data_waiting(self) -> bool:
        """Indicates if data is in the serial receive buffer."""
        return self.serial.in_waiting > 0
    
    def read_rx_buffer(self, timeout: int = 0, read_until: str = '') -> 'str|None':
        """Reads data from the serial receive buffer.
        
        Args:
            timeout (int): Optional timeout in seconds to wait for data.
                If 0, will stop reading when no character is waiting.
        
        Returns:
            The data string if any was present, else `None`.
            
        """
        if not isinstance(timeout, int):
            timeout = 0
        self.serial.flush()   # wait for anything sent prior to be done
        rx_data = ''
        start_time = time.time()
        while (time.time() - start_time < timeout or timeout == 0):
            while (self.serial.in_waiting > 0):
                rx_data += self._read()
            if timeout == 0 or rx_data.endswith(read_until):
                break
        if vlog(VLOG_TAG):
            if rx_data:
                _log.debug('Read from serial: %s', dprint(rx_data))
            else:
                _log.debug('No data waiting on serial buffer')
        return rx_data
    
    def send_at_command(self, at_command: str) -> None:
        """Submits an AT command to the NIMO modem to solicit a response.
        
        Use `read_at_response` then `get_resopnse` after sending the command.
        
        Args:
            at_command: The command to send.
        
        """
        if vlog(VLOG_TAG) and not self.ready.is_set():
            _log.debug('Waiting for prior command completion')
        self.ready.wait()
        self.ready.clear()
        if 'ATS80?' in at_command:
            _log.debug('Somebody is querying error state!')
        dump_buffer = self.read_rx_buffer()
        if dump_buffer:
            _log.warning('Dumping RX buffer: %s', dprint(dump_buffer))
        self._pending_command = at_command
        if self.crc and '*' not in at_command:
            self._pending_command = apply_crc(at_command)
        self._pending_command += '\r'
        if vlog(VLOG_TAG):
            _log.debug('Sending on serial: %s', dprint(self._pending_command))
        self.serial.write(self._pending_command.encode())
        self.serial.flush()   # ensure it gets sent
    
    def read_at_response(self,
                         prefix: str = None,
                         timeout: int = DEFAULT_AT_TIMEOUT,
                         tick: int = 0) -> AtErrorCode:
        """Parses the pending AT command response into a buffer.
        
        Use `send_at_command` prior to calling.
        Use `get_response` to retrieve the parsed response.
        
        Args:
            prefix: Optional prefix to remove from the response.
            timeout: Maximum time in seconds to wait for response (default 3)
            tick: Optional debug for timeout countdown in seconds
        
        Returns:
            Error code indicating success or reason for parsing error.
        
        Raises:
            `OSError` if there is no pending command.
        
        """
        if not isinstance(timeout, int):
            timeout = DEFAULT_AT_TIMEOUT
        if not self._pending_command:
            raise OSError('No pending command to read response for')
        if vlog(VLOG_TAG):
            _log.debug('Parsing response for %s', dprint(self._pending_command))
        self._rx_buffer = ''
        parsing = AtParsingState.ECHO if self.echo else AtParsingState.RESPONSE
        result_ok: bool = False
        crc_found: bool = False
        error: AtErrorCode = AtErrorCode.OK
        peeked: str = ''
        start_time = time.time()
        countdown = timeout
        while (time.time() - start_time < timeout and
               parsing < AtParsingState.OK):
            while ((self.serial.in_waiting > 0 or peeked) and
                   parsing < AtParsingState.OK):
                if peeked:
                    self._rx_buffer += peeked
                    peeked = ''
                else:
                    self._rx_buffer += self._read()
                last = self._rx_buffer[-1]
                if last == '\n':
                    if self._rx_buffer.endswith(VRES_OK):
                        result_ok = True
                        parsing = self._parsing_ok()
                    elif self._rx_buffer.endswith(VRES_ERR):
                        parsing = self._parsing_error()
                    elif parsing == AtParsingState.CRC:
                        if vlog(VLOG_TAG):
                            _log.debug('CRC parsing complete')
                        if not result_ok:
                            parsing = AtParsingState.ERROR
                        else:
                            if validate_crc(self._rx_buffer):
                                parsing = AtParsingState.OK
                            else:
                                _log.error('Invalid CRC')
                                parsing = AtParsingState.ERROR
                                error = AtErrorCode.INVALID_CRC
                                result_ok = False
                    # else response line terminator - keep parsing
                elif last == '\r':
                    if self._rx_buffer == self._pending_command:
                        if vlog(VLOG_TAG):
                            _log.debug('Echo received - clearing RX buffer')
                        self._rx_buffer = ''
                        parsing = AtParsingState.RESPONSE
                    else:
                        old_parsing = parsing
                        if self.serial.in_waiting == 0:
                            parsing = self._parsing_short(parsing)
                        else:
                            peeked = self._read()
                            if peeked == '*':
                                parsing = self._parsing_short()
                        if old_parsing != parsing:
                            result_ok = parsing == AtParsingState.OK
                else:
                    if parsing == AtParsingState.CRC and not crc_found:
                        if last == '*':
                            crc_found = True
                        else:
                            _log.warning('Unexpected CRC character %s', last)
            if parsing >= AtParsingState.OK:
                if vlog(VLOG_TAG):
                    _log.debug('Parsing complete')
                break
            if tick > 0 and self._rx_buffer == '':
                time.sleep(tick)
                countdown -= tick
                if vlog(VLOG_TAG):
                    _log.debug('Countdown: %d', countdown)
            time.sleep(self._char_delay)
        # end while < timeout
        if vlog(VLOG_TAG) and self._rx_buffer:
            _log.debug('Raw response: %s', dprint(self._rx_buffer))
        if parsing < AtParsingState.OK:
            _log.warning('AT command timeout during parsing')
            if self.verbose and self._rx_buffer.endswith('\r'):
                _log.info('Detected non-verbose - setting flag')
                self.verbose = False
            else:
                error = AtErrorCode.TIMEOUT
        elif parsing == AtParsingState.ERROR:
            error = AtErrorCode.ERROR
            if not self.crc and crc_found:
                _log.warning('CRC detected but not expected - setting flag')
                self.crc = True
                error = AtErrorCode.CRC_CONFIG_MISMATCH
            self._rx_buffer = ''
        else:
            if (self.crc):
                if vlog(VLOG_TAG):
                    _log.debug('Removing CRC')
                crc_length = 7
                if (self._rx_buffer[-crc_length:].startswith('*') and
                    self._rx_buffer[-crc_length:].endswith('\r\n')):
                    # CRC terminates response so remove it
                    self._rx_buffer = self._rx_buffer[:-crc_length]
                else:
                    _log.warning('CRC expected but not found - reset flag')
                    self.crc = False
            to_remove = VRES_OK if self.verbose else RES_OK
            self._rx_buffer = self._rx_buffer.replace(to_remove, '')
            if vlog(VLOG_TAG):
                _log.debug('Removed result code (%s): %s',
                           dprint(to_remove), dprint(self._rx_buffer))
            if prefix:
                self._rx_buffer = self._rx_buffer.replace(prefix, '', 1)
                if vlog(VLOG_TAG):
                    _log.debug('Removed prefix (%s): %s',
                               dprint(prefix), dprint(self._rx_buffer))
            self._rx_buffer = self._rx_buffer.strip()
            if vlog(VLOG_TAG):
                _log.debug('Trimmed leading/trailing whitespace: %s',
                           dprint(self._rx_buffer))
            self._rx_buffer = self._rx_buffer.replace('\r\n', '\n')
            self._rx_buffer = self._rx_buffer.replace('\n\n', '\n')
            if vlog(VLOG_TAG):
                _log.debug('Consolidated line feeds: %s',
                           dprint(self._rx_buffer))
        # cleanup
        self._pending_command = ''
        self.ready.set()
        return error
    
    def get_response(self) -> str:
        """Get the response following a read operation and clear the buffer."""
        response = self._rx_buffer
        self._rx_buffer = ''
        return response
    
    def read_line(self, read_until: str = '\r\n') -> str:
        """"""
        if vlog(VLOG_TAG) and not self.ready.is_set():
            _log.debug('Waiting for prior command/response completion')
        self.ready.wait()
        self.ready.clear()
        line: str = ''
        while self.serial.in_waiting > 0:
            line += self._read()
            if line.endswith(read_until) and len(line) > len(read_until):
                break
        self.ready.set()
        return line.strip()
    
    def _read(self) -> str:
        """Read an ASCII character or generate a warning."""
        char = self.serial.read()
        try:
            return char.decode()
        except UnicodeDecodeError as exc:
            _log.error('Unable to decode [%d] (%s)', char, exc)
            return ''
    
    def _parsing_ok(self) -> AtParsingState:
        """Internal helper for parsing valid response."""
        _log.debug('Result OK')
        if not self.crc:
            if 'CRC=1\r' in self._pending_command.upper():
                _log.debug('Command enabled CRC - set flag')
                self.crc = True
                return AtParsingState.CRC
        else:
            if 'CRC=0\r' in self._pending_command.upper():
                _log.debug('Command disabled CRC - reset flag')
                self.crc = False
            else:
                return AtParsingState.CRC
        return AtParsingState.OK
    
    def _parsing_error(self) -> AtParsingState:
        """Internal helper for parsing errored response."""
        _log.warning('Result ERROR')
        if self.crc or self.serial.in_waiting > 0:
            return AtParsingState.CRC
        else:
            time.sleep(self._char_delay)
            if self.serial.in_waiting > 0:
                return AtParsingState.CRC
        return AtParsingState.ERROR
    
    def _parsing_short(self, current: AtParsingState = None) -> AtParsingState:
        """Internal helper for parsing short code responses."""
        if vlog(VLOG_TAG):
            _log.debug('Check short response for: %s', dprint(self._rx_buffer))
        if (not self._rx_buffer.endswith((RES_OK, RES_ERR)) or
            (self.verbose and self._rx_buffer.startswith('\r\n'))):
            # just reading too fast, wait for next character
            return current
        if self.verbose:
            _log.warning('Clearing verbose flag due to short response match')
            self.verbose = False
        if self._rx_buffer.endswith(RES_OK):
            return self._parsing_ok()
        return self._parsing_error()



"""Library to interface with a Viasat-approved NIMO modem for satellite IoT.

This library abstracts various low-level AT command operations useful for
interacting with a NIMO modem to send and receive data, check network status
and get location-based information.

Most `get` methods will raise a `ModemAtError` if a valid response is not
received to a command/query.

`ModemTimeout` will be raised if no response is received to a command within
the default or specified timeout.

AT command errors will raise `ModemAtError` with a property `error_code` to
provide further details with the `AtErrorCode`.

"""

from .constants import (
    AtErrorCode,
    BeamState,
    ControlState,
    DataFormat,
    EventNotification,
    GnssMode,
    MessagePriority,
    MessageState,
    NetworkStatus,
    PowerMode,
    SignalQuality,
    UrcCode,
    UrcControl,
    WakeupPeriod,
    WakeupWay,
    WorkMode,
)
from .modem import (
    Manufacturer,
    ModemLocation,
    MoMessage,
    MtMessage,
    NimoModem,
    ModemAtError,
    ModemCrcConfig,
    ModemError,
    ModemTimeout,
    AcquisitionInfo,
    SatelliteLocation,
)

__all__ = [
    'AtErrorCode',
    'BeamState',
    'ControlState',
    'DataFormat',
    'GnssMode',
    'Manufacturer',
    'MessagePriority',
    'MessageState',
    'ModemLocation',
    'MoMessage',
    'MtMessage',
    'NetworkStatus',
    'NimoModem',
    'ModemAtError',
    'ModemCrcConfig',
    'ModemError',
    'ModemTimeout',
    'PowerMode',
    'AcquisitionInfo',
    'SatelliteLocation',
    'SignalQuality',
    'WakeupPeriod',
    'WakeupWay',
    'WorkMode',
    'UrcCode',
    'UrcControl',
    'EventNotification',
]


"""IDP modem S-register definitions

This module provides NIMO/IDP modem S-register definitions useful for
creating a digital twin of the modem.
Updated register map of default and configured values can be read from the modem
using AT%SREG which returns a human-readable table.

"""
# Tuples: (name[0], default[1], read-only[2], range[3], description[4], note[5])
ORBCOMM_REGISTER_DEFINITIONS = [
    ('S0', 0, True, [0, 255], 'auto answer', 'unused'),
    ('S3', 13, False, [1, 127], 'command termination character', None),
    ('S4', 10, False, [0, 127], 'response formatting character', None),
    ('S5', 8, False, [0, 127], 'command line editing character', None),
    ('S6', 0, True, [0, 255], 'pause before dial', 'unused'),
    ('S7', 0, True, [0, 255], 'connection completion timeout', 'unused'),
    ('S8', 0, True, [0, 255], 'commia dial modifier time', 'unused'),
    ('S10', 0, True, [0, 255], 'automatic discovery delay', 'unused'),
    ('S31', 80, False, [10, 250], 'DOP threshold (x10)', None),
    ('S32', 25, False, [1, 1000], 'position accuracy threshold [m]', None),
    ('S33', 0, False, [0, 8], 'default dynamic platform model', None),
    ('S34', 7, True, [0, 255], 'Doppler dynamic platform model', 'Reserved'),
    ('S35', 0, False, [0, 255], 'static hold threshold [cm/s]', None),
    ('S36', 0, False, [-1, 480], 'standby timeout [min]', None),
    ('S37', 200, False, [1, 1000], 'speed accuracy threshold', None),
    ('S38', 1, True, [0, 0], 'reserved', None),
    ('S39', 0, False, [0, 2], 'GNSS mode', None),
    ('S40', 0, False, [0, 60], 'GNSS signal satellite detection timeout', None),
    ('S41', 180, False, [60, 1200], 'GNSS fix timeout', None),
    ('S42', 65535, False, [0, 65535], 'GNSS augmentation systems', 'Query fails'),
    ('S50', 0, False, [0, 9], 'power mode', None),
    ('S51', 0, False, [0, 6], 'wakeup interval', None),
    ('S52', 2500, True, [0, 2500], 'reserved', 'undocumented'),
    ('S53', 0, True, [0, 255], 'satcom control', None),
    ('S54', 0, True, [0, 0], 'satcom status', None),
    ('S55', 0, False, [0, 30], 'GNSS continuous mode', None),
    ('S56', 0, True, [0, 255], 'GNSS jamming status', None),
    ('S57', 0, True, [0, 255], 'GNSS jamming indicator', None),
    ('S60', 1, False, [0, 1], 'Echo', None),
    ('S61', 0, False, [0, 1], 'Quiet', None),
    ('S62', 1, False, [0, 1], 'Verbose', None),
    ('S63', 0, False, [0, 1], 'CRC', None),
    ('S64', 42, False, [0, 255], 'prefix character of CRC sequence', None),
    ('S70', 0, True, [0, 0], 'reserved', 'undocumented'),
    ('S71', 0, True, [0, 0], 'reserved', 'undocumented'),
    ('S80', 0, True, [0, 255], 'last error code', None),
    ('S81', 0, True, [0, 255], 'most recent result code', None),
    ('S85', 22, True, [0, 0], 'temperature', None),
    ('S88', 0, False, [0, 65535], 'event notification control', None),
    # ('S89', 0, False, [0, 65535], 'event notification status', None),
    ('S90', 0, False, [0, 7], 'capture trace define - class', None),
    ('S91', 0, False, [0, 31], 'capture trace define - subclass', None),
    ('S92', 0, False, [0, 255], 'capture trace define - initiate', None),
    ('S93', 0, True, [0, 255], 'captured trace property - data size', None),
    ('S94', 0, True, [0, 255], 'captured trace property - signed indicator', None),
    ('S95', 0, True, [0, 255], 'captured trace property - mobile ID', None),
    ('S96', 0, True, [0, 255], 'captured trace property - timestamp', None),
    ('S97', 0, True, [0, 255], 'captured trace property - class', None),
    ('S98', 0, True, [0, 255], 'captured trace property - subclass', None),
    ('S99', 0, True, [0, 255], 'captured trace property - severity', None),
    ('S100', 0, True, [0, 255], 'captured trace data 0', None),
    ('S101', 0, True, [0, 255], 'captured trace data 1', None),
    ('S102', 0, True, [0, 255], 'captured trace data 2', None),
    ('S103', 0, True, [0, 255], 'captured trace data 3', None),
    ('S104', 0, True, [0, 255], 'captured trace data 4', None),
    ('S105', 0, True, [0, 255], 'captured trace data 5', None),
    ('S106', 0, True, [0, 255], 'captured trace data 6', None),
    ('S107', 0, True, [0, 255], 'captured trace data 7', None),
    ('S108', 0, True, [0, 255], 'captured trace data 8', None),
    ('S109', 0, True, [0, 255], 'captured trace data 9', None),
    ('S110', 0, True, [0, 255], 'captured trace data 10', None),
    ('S111', 0, True, [0, 255], 'captured trace data 11', None),
    ('S112', 0, True, [0, 255], 'captured trace data 12', None),
    ('S113', 0, True, [0, 255], 'captured trace data 13', None),
    ('S114', 0, True, [0, 255], 'captured trace data 14', None),
    ('S115', 0, True, [0, 255], 'captured trace data 15', None),
    ('S116', 0, True, [0, 255], 'captured trace data 16', None),
    ('S117', 0, True, [0, 255], 'captured trace data 17', None),
    ('S118', 0, True, [0, 255], 'captured trace data 18', None),
    ('S119', 0, True, [0, 255], 'captured trace data 19', None),
    ('S120', 0, True, [0, 255], 'captured trace data 20', None),
    ('S121', 0, True, [0, 255], 'captured trace data 21', None),
    ('S122', 0, True, [0, 255], 'captured trace data 22', None),
    ('S123', 0, True, [0, 255], 'captured trace data 23', None),
]


class SRegister(object):
    """
    Twin of a modem S register
    """

    def __init__(self,
                 name: str,
                 default: int,
                 read_only: bool,
                 min: int,
                 max: int,
                 description: str,
                 note: str = None):
        """
        Initializes an S register twin

        :param name: (string) name of the S register e.g. 'S50'
        :param default: (int) default value of the register
        :param read_only: (Boolean)
        :param low: (int) lowest value allowed
        :param high: (int) highest value allowed
        :param description: (string)
        :param note: (string), defaults to None
        """
        self._name = name
        self._default = default
        self._value = default
        self._read_only = read_only
        self._min = min
        self._max = max
        self._description = description
        self._note = note

    @property
    def name(self) -> str:
        return self._name

    @property
    def default(self) -> int:
        return self._default

    @property
    def value(self) -> int:
        return self._value

    @value.setter
    def value(self, value: int):
        if self.read_only:
            raise ValueError(f'{self.name} is read-only')
        if not isinstance(value, int) or value not in self.min_max:
            raise ValueError(f'{self.name} must be {self.min}..{self.max}')
        self._value = value

    @property
    def read_only(self) -> bool:
        return self._read_only

    @property
    def min_max(self) -> range:
        return range(self._min, self._max +1)

    @property
    def min(self) -> int:
        return self._min
    
    @property
    def max(self) -> int:
        return self._max

    @property
    def description(self) -> str:
        return self._description

    @property
    def note(self) -> str:
        return self._note


class SRegisters(dict):
    """A dictionary twin of the modem's S registers."""
    def __init__(self):
        for tup in ORBCOMM_REGISTER_DEFINITIONS:
            name, dflt, ro, lo_hi, desc, note = tup
            self[name] = SRegister(name,
                                   dflt,
                                   ro,
                                   lo_hi[0],
                                   lo_hi[1],
                                   desc,
                                   note)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pynimomodem']

package_data = \
{'': ['*']}

install_requires = \
['pyserial>=3.5,<4.0']

setup_kwargs = {
    'name': 'pynimomodem',
    'version': '0.4.3',
    'description': "A Python implementation of Viasat's NIMO modem interface project.",
    'long_description': "# pynimomodem\n\nA Python implementation of the [Viasat](www.viasat.com)\nNIMO modem interface for satellite IoT.\n\nNIMO stands for **Non-IP Modem Orbcomm** waveform\nand represents a family of low cost satellite data modems that use network\nprotocols developed by [ORBCOMM](www.orbcomm.com)\nincluding [IsatData Pro](https://www.inmarsat.com/en/solutions-services/enterprise/services/isatdata-pro.html) and its successor, OGx.\n\nThese ORBCOMM protocols can operate over the Viasat L-band global network in\ncooperation with a varietry of authorized Viasat IoT service partners, and\nare intended for event-based remote data collection and device control.\n\nExample NIMO modems available:\n* [ORBCOMM ST2100](https://www.orbcomm.com/en/partners/iot-hardware/st-2100)\n* [Quectel CC200A-LB](https://www.quectel.com/product/cc200a-lb-satellite-communication-module)\n* [uBlox UBX-S52](https://content.u-blox.com/sites/default/files/documents/UBX-R52-S52_ProductSummary_UBX-19026227.pdf)\n\n## Additional Information\n\n* [API Documentation](https://inmarsat-enterprise.github.io/pynimomodem/)\n\n> [!NOTE]\n> Obsoletes/replaces the Inmarsat `idpmodem` project, when combined with the\n> [`pynimcodec`](github.com/inmarsat-enterprise/pynimcodec) library.\n\n## Installation\n\nExample using pip, on a Linux-based platform including `PySerial` dependency:\n```\npip install 'pynimomodem'\n```\n\n## Background\n\n### Overview\n\n*IsatData Pro* (**IDP**)is a store-and-forward satellite messaging technology\nwith flexible message sizes:\n\n* up to 6400 bytes Mobile-Originated (aka **MO**, From-Terminal, *Return*)\n* up to 10000 bytes Mobile-Terminated (aka **MT**, To-Terminal, *Forward*)\n\n***Message***s are sent to or collected from a ***Mobile*** using its globally\nunique *Mobile ID*,\ntransacted through a ***Mailbox*** that provides authentication, encryption and\ndata segregation for cloud-based or enterprise client applications via a\nREST **Messaging API**.\n\nData sources and controls in the field are interfaced to a modem using a serial\ninterface with *AT commands* to send and receive messages, check network status,\nand optionally use the built-in *Global Navigation Satellite System* (GNSS)\nreceiver to determine location-based information.\n\nThe first byte of the message is referred to as the\n*Service Identification Number* (**SIN**) where values below 16 are reserved\nfor system use.  SIN is intended to capture the concept of embedded\nmicroservices used by an application.\n\nThe second byte of the message can optionally be defined as the\n*Message Identifier Number* (**MIN**) intended to support remote operations \nwithin each embedded microservice with defined binary formatting.\nThe MIN concept also supports the optional *Message Definition File* feature\nallowing an XML file to be applied which presents a JSON-tagged message\nstructure on the network API.\n\n### Modem Concept of Operation\n\n1. Upon power-up or reset, the modem first acquires its location using \nGlobal Navigation Satellite Systems (GNSS).\n1. After getting its location, the modem tunes to the correct frequency, then\nregisters on the network.  Once registered it can communicate on the\nnetwork.\n1. MO messages are submitted by a microcontroller or IoT Edge device, which\nthen must monitor progress until the message is complete (either delivered or\ntimed out/failed due to blockage). Completed messages must be cleared from the\nmodem transmit queue by querying state(s) either periodically or when prompted\nby the modem's event notification pin if configured.\n1. MT messages that arrive are stored in the receive queue and the Edge device\nqueries for *New* MT messages periodically or when prompted by the modem's\nevent notification pin if configured.\n1. Network acquisition status can also be queried using AT commands.\n1. If the modem cannot find the target frequency it begins to search for other\nfrequencies from a configuration map in its non-volatile memory. It will cycle\nthrough beam acquisition attempts for a period of time before falling back to\na globally-accessible *Bulletin Board* frequency where it may need to download\na new network configuration before re-attempting. Bulletin board downloads\ntypically take less than 15 minutes but can take longer in low signal or high\ninterference locations. A modem should not be powered off during Bulletin\nBoard download.\n1. Prolonged obstruction of satellite signal will put the modem into *blockage*\nstate from which it will automatically try to recover based on an algorithm\ninfluenced by its *power mode* setting.",
    'author': 'geoffbrucepayne',
    'author_email': 'geoff.bruce-payne@inmarsat.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/inmarsat-enterprise/pynimomodem',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Classes and methods for location, elevation and azimuth for NIMO modems.

This module:

* Parses NMEA-0183 data into a `ModemLocation` object.
* Calculates azimuth and elevation to a geostationary `SatelliteLocation`.

Thanks for Azimuth/Elevation derived from code at:
https://github.com/sq3tle/altazrange/tree/master

"""
import json
import logging
import math
from copy import deepcopy
from dataclasses import dataclass

from .constants import (
    GEOSTATIONARY_DISTANCE_M,
    GeoBeam,
    GeoSatellite,
    NimoIntEnum,
)
from .nimoutils import iso_to_ts, ts_to_iso, vlog

VLOG_TAG = 'nmealocation'
TRACE_TAG = VLOG_TAG + 'trace'

_log = logging.getLogger(__name__)


class GnssFixType(NimoIntEnum):
    """Enumerated fix type from NMEA-0183 standard."""
    NONE = 1
    D2 = 2
    D3 = 3


class GnssFixQuality(NimoIntEnum):
    """Enumerated fix quality from NMEA-0183 standard."""
    INVALID = 0
    GPS_SPS = 1
    DGPS = 2
    PPS = 3
    RTK = 4
    FLOAT_RTK = 5
    EST_DEAD_RECKONING = 6
    MANUAL = 7
    SIMULATION = 8


@dataclass
class GnssSatelliteInfo(object):
    """Information specific to a GNSS satellite.
    
    Attributes:
        prn: The PRN code (Pseudo-Random Number sequence)
        elevation: The satellite elevation
        azimuth: The satellite azimuth
        snr: The satellite Signal-to-Noise Ratio
    """
    prn: int
    elevation: int
    azimuth: int
    snr: int


class ModemLocation:
    """A set of location-based information derived from the modem's NMEA data.
    
    Uses 90.0/180.0 if latitude/longitude are unknown

    Attributes:
        latitude (float): decimal degrees
        longitude (float): decimal degrees
        altitude (float): in metres
        speed (float): in knots
        heading (float): in degrees
        timestamp (int): in seconds since 1970-01-01T00:00:00Z
        satellites (int): in view at time of fix
        fix_type (GnssFixType): 1=None, 2=2D or 3=3D
        fix_quality (GnssFixQuality): Enumerated lookup value
        pdop (float): Probability Dilution of Precision
        hdop (float): Horizontal Dilution of Precision
        vdop (float): Vertical Dilution of Precision
        time_iso (str): ISO 8601 formatted timestamp

    """
    def __init__(self, **kwargs):
        """Initializes a Location with default latitude/longitude 90/180."""
        self.latitude = float(kwargs.get('latitude', 90.0))
        self.longitude = float(kwargs.get('longitude', 180.0))
        self.altitude = float(kwargs.get('altitude', 0.0))   # metres
        self.speed = float(kwargs.get('speed', 0.0))  # knots
        self.heading = float(kwargs.get('heading', 0.0))   # degrees
        self.timestamp = int(kwargs.get('timestamp', 0))   # seconds (unix)
        self.satellites = int(kwargs.get('satellites', 0))
        self.fix_type = GnssFixType(int(kwargs.get('fix_type', 1)))
        self.fix_quality = GnssFixQuality(int(kwargs.get('fix_quality', 0)))
        self.pdop = float(kwargs.get('pdop', 99))
        self.hdop = float(kwargs.get('hdop', 99))
        self.vdop = float(kwargs.get('vdop', 99))
        # self.satellites_info: 'list[GnssSatelliteInfo]' = kwargs.get(
        #     'satellites_info', []
        # )

    @property
    def time_iso(self) -> str:
        return f'{ts_to_iso(self.timestamp)}'

    # def _update_satellites_info(self,
    #                             satellites_info: 'list[GnssSatelliteInfo]'):
    #     """Populates satellite information based on NMEA GSV data."""
    #     for satellite_info in satellites_info:
    #         if isinstance(satellite_info, GnssSatelliteInfo):
    #             new = True
    #             for i, info in enumerate(self.satellites_info):
    #                 if info.prn == satellite_info.prn:
    #                     new = False
    #                     self.satellites_info[i] = satellite_info
    #                     break
    #             if new:
    #                 self.satellites_info.append(satellite_info)

    def __repr__(self) -> str:
        obj = deepcopy(self.__dict__)
        for k, v in obj.items():
            if k in ['latitude', 'longitude']:
                obj[k] = round(v, 5)
            elif isinstance(v, float):
                obj[k] = round(v, 1)
        return json.dumps(obj, skipkeys=True)


@dataclass
class SatelliteLocation:
    """Represents a geostationary satellite location relative to a modem."""
    name: str = ''
    latitude: float = 0.0
    longitude: float = 180.0
    altitude: float = GEOSTATIONARY_DISTANCE_M
    azimuth: float = 0.0
    elevation: float = 0.0
    geobeam: 'GeoBeam|None' = None


@dataclass
class _PointLocation:
    """Location used for spherical coordinates (for azimuth/elevation)."""
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0
    radius: float = 0.0
    nx: float = 0.0
    ny: float = 0.0
    nz: float = 0.0


def validate_nmea(nmea_sentence: str) -> bool:
    """Validates a given NMEA-0183 sentence with CRC.
    
    Args:
        nmea_sentence (str): NMEA-0183 sentence ending in checksum.
    
    """
    if '*' not in nmea_sentence:
        return False
    data, cs_hex = nmea_sentence.split('*')
    candidate = int(cs_hex, 16)
    crc = 0   # initial
    for i in range(1, len(data)):   # ignore initial $
        crc ^= ord(data[i])
    return candidate == crc


def parse_nmea_to_location(location: ModemLocation, nmea_sentence: str) -> None:
    """Parses a NMEA-0183 sentence to update a ModemLocation."""
    if vlog(VLOG_TAG):
        _log.debug('Parsing NMEA: %s', nmea_sentence)
    if not validate_nmea(nmea_sentence):
        raise ValueError('Invalid NMEA-0183 sentence')
    data = nmea_sentence.split('*')[0]
    nmea_type = ''
    cache = {}
    for i, field_data in enumerate(data.split(',')):
        if i == 0:
            nmea_type = field_data[-3:]
            if nmea_type == 'GSV':
                _log.warn('No processing required for GSV sentence')
                return
            if nmea_type == 'GSA' and location.vdop != 99:
                if vlog(TRACE_TAG):
                    _log.debug('Skipping redundant GSA data')
                return
            if vlog(TRACE_TAG):
                _log.debug('Processing NMEA type: %s', nmea_type)
        elif i == 1:
            if nmea_type == 'RMC':
                cache['fix_hour'] = field_data[0:2]
                cache['fix_min'] = field_data[2:4]
                cache['fix_sec'] = field_data[4:6]
                if vlog(TRACE_TAG):
                    _log.debug('Fix time %s:%s:%s', cache['fix_hour'],
                               cache['fix_min'], cache['fix_sec'])
        elif i == 2:
            if nmea_type == 'RMC':
                if (field_data == 'V'):
                    _log.warn('Fix Void')
            elif nmea_type == 'GSA':
                location.fix_type = GnssFixType(int(field_data))
                if vlog(TRACE_TAG):
                    _log.debug('Fix type: %s', location.fix_type.name)
        elif i == 3:
            if nmea_type == 'RMC':
                location.latitude = (float(field_data[0:2]) +
                                     float(field_data[2]) / 60.0)
        elif i == 4:
            if nmea_type == 'RMC':
                if field_data == 'S':
                    location.latitude *= -1
                if vlog(TRACE_TAG):
                    _log.debug('Latitude: %.5f', location.latitude)
        elif i == 5:
            if nmea_type == 'RMC':
                location.longitude = (float(field_data[0:3]) +
                                      float(field_data[3]) / 60.0)
        elif i == 6:
            if nmea_type == 'RMC':
                if field_data == 'W':
                    location.longitude *= -1
                if vlog(TRACE_TAG):
                    _log.debug('Longitude: %.5f', location.longitude)
            elif nmea_type == 'GGA':
                location.fix_quality = GnssFixQuality(int(field_data))
                if vlog(TRACE_TAG):
                    _log.debug('Fix quality: %s', location.fix_quality.name)
        elif i == 7:
            if nmea_type == 'RMC':
                location.speed = float(field_data)
                if vlog(TRACE_TAG):
                    _log.debug('Speed: %.1f', location.speed)
            elif nmea_type == 'GGA':
                location.satellites = int(field_data)
                if vlog(TRACE_TAG):
                    _log.debug('GNSS satellites used: %d', location.satellites)
        elif i == 8:
            if nmea_type == 'RMC':
                location.heading = float(field_data)
                if vlog(TRACE_TAG):
                    _log.debug('Heading: %.1f', location.heading)
            elif nmea_type == 'GGA':
                location.hdop = round(float(field_data), 1)
                if vlog(TRACE_TAG):
                    _log.debug('HDOP: %.1f', location.heading)
        elif i == 9:
            if nmea_type == 'RMC':
                fix_day = field_data[0:2]
                fix_month = field_data[2:4]
                fix_yy = int(field_data[4:])
                fix_yy += 1900 if fix_yy >= 73 else 2000
                if vlog(TRACE_TAG):
                    _log.debug('Fix date %d-%s-%s', fix_yy, fix_month, fix_day)
                iso_time = (f'{fix_yy}-{fix_month}-{fix_day}T'
                            f'{cache["fix_hour"]}:{cache["fix_min"]}'
                            f':{cache["fix_sec"]}Z')
                unix_timestamp = iso_to_ts(iso_time)
                if vlog(TRACE_TAG):
                    _log.debug('Fix time ISO 8601: %s | Unix: %d',
                               iso_time, unix_timestamp)
                location.timestamp = unix_timestamp
            elif nmea_type == 'GGA':
                location.altitude = float(field_data)
                if vlog(TRACE_TAG):
                    _log.debug('Altitude: %.1f', location.altitude)
        elif i == 10:
            # RMC magnetic variation - ignore
            if nmea_type == 'GGA' and field_data != 'M':
                _log.warning('Unexpected altitude units: %s', field_data)
        # elif i == 11:   # RMC magnetic variation direction, GGA height of geoid - ignore
        # elif i == 12:   # GGA units height of geoid - ignore
        # elif i == 13:   # GGA seconds since last DGPS update - ignore
        # elif i == 14:   # GGA DGPS station ID - ignore
        elif i == 15:   # GSA PDOP - ignore (unused)
            if nmea_type == 'GSA':
                location.pdop = round(float(field_data), 1)
                if vlog(TRACE_TAG):
                    _log.debug('PDOP: %d', location.pdop)
        # elif i == 16:   # GSA HDOP - ignore (use GGA)
        elif i == 17:
            if nmea_type == 'GSA':
                location.vdop = round(float(field_data), 1)
                if vlog(TRACE_TAG):
                    _log.debug('VDOP: %d', location.vdop)


def get_location_from_nmea_data(nmea_data: 'str|list[str]') -> ModemLocation:
    """Derives a ModemLocation from a set of NMEA-0183 sentences.
    
    Args:
        nmea_data (str): A set of NMEA-0183 sentences separated by `\n` or
            a `list` of NMEA sentences.
    
    Returns:
        `Location` object.
    
    """
    location = ModemLocation()
    if isinstance(nmea_data, list):
            if not all(isinstance(x, str) for x in nmea_data):
                raise ValueError('Invalid NMEA sentence list')
    elif isinstance(nmea_data, str):
        nmea_data = nmea_data.split('\n')
    for nmea_sentence in nmea_data:
        parse_nmea_to_location(location, nmea_sentence)
    if vlog(VLOG_TAG):
        _log.debug('Location: %s', repr(location))
    return location


def get_closest_satellite(latitude: float, longitude: float) -> GeoSatellite:
    """Get the closest geostationary satellite to a given location."""
    satellites = list(map(lambda x: x.value,GeoSatellite._member_map_.values()))
    closest = GeoSatellite(min(satellites, key=lambda x:abs(x-longitude)))
    if closest == GeoSatellite.AORWSC:   #: single regional beam only
        if latitude >= 15.0 or latitude <= -45.0:
            if longitude >= -27.0:
                return GeoSatellite.EMEA
            return GeoSatellite.AMER
    # elif closest == GeoSatellite.MEAS:   #: not all beams lit
    #     if latitude <= -4.5:
    #         if longitude >= 63.5:
    #             return GeoSatellite.APAC
    #         return GeoSatellite.EMEA
    #     elif latitude >= 40.9:
    #         if longitude <= 45.0:
    #             return GeoSatellite.EMEA
    #         if longitude >= 82.5:
    #             return GeoSatellite.APAC
    #     elif longitude >= 63.5:
    #         if latitude <= -4.0 or latitude >= 30.0:
    #             return GeoSatellite.APAC
    return closest


def get_satellite_location(modem_location: ModemLocation,
                           geobeam: 'GeoBeam|None' = None,
                           ) -> SatelliteLocation:
    """Derives the azimuth and elevation of the nearest satellite.
    
    If not provided the current GeoBeam, derives the closest satellite from
    the location provided.
    
    Args:
        modem_location (ModemLocation): The modem's Location object.
        geobeam (GeoBeam): The GeoBeam, if known. If `None` it will be
            assumed from fixed locations of satellites.
    
    Returns:
        `SatelliteLocation` including azimuth, elevation from the modem.
    
    """
    # internal helper functions
    def location_to_point(latitude: float,
                        longitude: float,
                        altitude: float,
                        ) -> _PointLocation:
        """Converts lat/lon/alt to point location."""
        lat_rad = latitude * math.pi / 180.0
        lon_rad = longitude * math.pi / 180.0
        radius = get_geographic_radius(lat_rad)
        clat = get_geocentric_latitude(lat_rad)
        cos_lon = math.cos(lon_rad)
        sin_lon = math.sin(lon_rad)
        cos_lat = math.cos(clat)
        sin_lat = math.sin(clat)
        x = radius * cos_lon * cos_lat
        y = radius * sin_lon * cos_lat
        z = radius * sin_lat
        cos_glat = math.cos(lat_rad)
        sin_glat = math.sin(lat_rad)
        nx = cos_glat * cos_lon
        ny = cos_glat * sin_lon
        nz = sin_glat
        x += altitude * nx
        y += altitude * ny
        z += altitude * nz
        return _PointLocation(x, y, z, radius, nx, ny, nz)

    def get_geographic_radius(lat_rad: float) -> float:
        """Adjust radius for earth shape."""
        equatorial = 6378137.0
        polar = 6356752.3
        cos = math.cos(lat_rad)
        sin = math.sin(lat_rad)
        t1 = equatorial**2 * cos
        t2 = polar**2 * sin
        t3 = equatorial * cos
        t4 = polar * sin
        return math.sqrt((t1 * t1 + t2 * t2) / (t3 * t3 + t4 * t4))

    def get_geocentric_latitude(lat_rad: float) -> float:
        """Derives the geocentric latitude."""
        e2 = 0.00669437999014   # first eccentricity squared, constant
        clat = math.atan((1.0 - e2) * math.tan(lat_rad))
        return clat

    def rotate_globe(b: SatelliteLocation, a: ModemLocation, b_radius: float):
        """Rotate the globe for vector alignment."""
        brp = location_to_point(b.latitude, b.longitude - a.longitude, b.altitude)
        alat = get_geocentric_latitude(-a.latitude * math.pi / 180.0)
        acos = math.cos(alat)
        asin = math.sin(alat)
        bx = (brp.x * acos) - (brp.z * asin)
        by = brp.y
        bz = (brp.x * asin) + (brp.z * acos)
        return _PointLocation(bx, by, bz, b_radius, 0.0, 0.0, 0.0)

    def normalize_vector(b: _PointLocation, a: _PointLocation):
        """Normalize the difference between vectors."""
        dx = b.x - a.x
        dy = b.y - a.y
        dz = b.z - a.z
        dist2 = dx**2 + dy**2 +dz**2
        if dist2 == 0:
            return None
        dist = math.sqrt(dist2)
        return _PointLocation(dx/dist, dy/dist, dz/dist, 1.0, 0.0, 0.0, 0.0)

    azimuth = None
    elevation = None
    if isinstance(geobeam, GeoBeam):
        satellite_name = geobeam.name.split('_')[0]
        satellite = GeoSatellite[satellite_name]
    else:
        satellite = get_closest_satellite(modem_location.latitude,
                                          modem_location.longitude)
    # modem and satellite Cartesian location
    mc = modem_location
    sc = SatelliteLocation(name=satellite.name,
                           longitude=satellite.value, 
                           altitude=GEOSTATIONARY_DISTANCE_M,
                           geobeam=geobeam)
    # modem and satellite Point location
    mp = location_to_point(mc.latitude, mc.longitude, mc.altitude)
    sp = location_to_point(sc.latitude, sc.longitude, sc.altitude)
    ref_point = rotate_globe(sc, mc, sp.radius)
    if ref_point.z**2 + ref_point.y**2 > 1.0e-6:
        theta = math.atan2(ref_point.z, ref_point.y) * 180.0 / math.pi
        azimuth = 90.0 - theta
        if azimuth < 0.0:
            azimuth += 360.0
        if azimuth > 360.0:
            azimuth -= 360.0
        bma = normalize_vector(sp, mp)   # Bayesian model averaging
        if isinstance(bma, _PointLocation):
            elevation = (90.0 - (180.0 / math.pi) * math.acos(
                bma.x * mp.nx + bma.y * mp.ny + bma.z * mp.nz))
    sc.azimuth = round(azimuth, 1)
    sc.elevation = round(elevation, 1)
    return sc

# Below is parked for potential future use

# def _parse_gsv_to_location(location: Location, gsv_sentence: str) -> None:
#     """Returns a Location object based on an NMEA sentences data set.
#
#     Placeholder - overcomplicates Location object
#
#     Args:
#         location: The Location object to update
#         gsv_sentence: The Satellites in View sentence to parse
#
#     """
#     update_satellites = getattr(location, '_update_satellites_info', None)
#     if not callable(update_satellites):
#         raise ValueError('Location object does not support GSV parsing')
#     gsv = gsv_sentence.split(',')       # $GPGSV,2,1,08,01,40,083,46,02,17,308,41,12,07,344,39,14,22,228,45*75
#     '''
#     gsv_sentences = gsv[1]           # Number of sentences for full data
#     gsv_sentence = gsv[2]            # Sentence number (up to 4 satellites per sentence)
#     '''
#     gsv_satellites = gsv[3]          # Number of satellites in view
#     # following supports up to 4 satellites per sentence
#     satellites_info = []
#     if (len(gsv) - 4) % 4 > 0:
#         # TODO: warn/log this case of extra GSV data in sentence
#         pass
#     num_satellites_in_sentence = int((len(gsv)-4)/4)
#     for i in range(1, num_satellites_in_sentence+1):
#         prn = int(gsv[i*4]) if gsv[i*4] != '' else 0             # satellite PRN number
#         elevation = int(gsv[i*4+1]) if gsv[i*4+1] != '' else 0   # Elevation in degrees
#         azimuth = int(gsv[i*4+2]) if gsv[i*4+2] != '' else 0     # Azimuth in degrees
#         snr = int(gsv[i*4+3]) if gsv[i*4+3] != '' else 0         # Signal to Noise Ratio
#         satellites_info.append(GnssSatelliteInfo(prn,
#                                                     elevation,
#                                                     azimuth,
#                                                     snr))
#     location._update_satellites_info(satellites_info)
#     satellites = int(gsv_satellites) if gsv_satellites != '' else 0
#     if location.satellites < satellites:
#         location.satellites = satellites
#     else:
#         # TODO: log this case; should be limited to GPS simulation in Modem Simulator (3 satellites)
#         pass
